#ifndef lint
static char rcsid[] =
	"@(#)$Header: hsis.c,v 1.10 91/11/26 23:05:14 clynn Rel $";
static char copyright[] =
	"Copyright (c) 1990 Regents of the University of California";
#endif

/*
 * Copyright (c) 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Lawrence Berkeley Laboratory,
 * Berkeley, CA.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "hsis.h"
#if NHSIS > 0
#include <sys/param.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/errno.h>
#include <sys/systm.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <net/netisr.h>
#include <sys/stream.h>
#include <sys/ttycom.h>
#include <sys/tty.h>
#include <sys/time.h>

#include <sys/sockio.h>
#include <sundev/zsreg.h>
#include <sundev/mbvar.h>
#include <sun4c/intreg.h>

#include <sun4c/auxio.h>

#include "bpfilter.h"
#if NBPFILTER > 0
#include <net/bpf.h>
#endif

#include "z16c35.h"
#include "hsisreg.h"
#include "syncmode.h"
#include "syncstat.h"
#include "hsiscom.h"
#include "ppp.h"

#ifndef AF_RAWSYNC
#define AF_RAWSYNC AF_IMPLINK
#endif

int hsidebug = 0;
#define dprintf(x)	if(hsidebug)printf x;

/*
 * Like most things that have to deal with Zilog chips, this driver
 * requires two interrupt levels: a hardware level that should be as
 * high as possible (SBus level 5-7) and a software level that should
 * be at or below splimp (i.e., at level 4 or 6 on a sparc).  Very little
 * time is spent at hardware interrupt level.  Packet copies to and
 * from the board (<300 us, worst case) and almost all the device and
 * system stuff (20-100 us typically, 300 us if we have to call Sun's
 * incredibly slow Streams NIT) is done at the software interrupt level.
 */
#define splboard spl4
#define HARDINT_LEVEL 5

/*
 * Software interrupt level, state and status bits.
 */
#define SOFTINT_LEVEL 4
#define splsoft spl2

u_int hsis_isum;
#define RECV_DONE 1
#define XMIT_DONE 2
#define SOFTINT(softc, event) (hsis_isum |= (event) << \
						((sc)->sc_if.if_unit << 1))

/*
 * default set-up for scc (possibly modified by user-specified `syncmode').
 * see hsis_init for how this is interpreted.  Each entry is a particular
 * scc reg.  The scc requires register be set in a particular order and
 * not all bits (in particular, the rx & tx enables) can be loaded from
 * here (see hsis_init).  Be careful when changing this array.
 */
u_char hsis_scc_setup[] = {
	/* 0 */	0,
#ifdef HSIS_EXTERNAL_RCVDONE
	/* 1 */	ZSWR1_SIE,
#else
	/* 1 */	ZSWR1_SIE | ZSWR1_RIE_SPECIAL_ONLY,
#endif
	/* 2 */	0,
	/* 3 */	ZSWR3_RXCRC_ENABLE | ZSWR3_RX_8,
	/* 4 */	ZSWR4_SDLC,
	/* 5 */	ZSWR5_TXCRC_ENABLE | ZSWR5_RTS | ZSWR5_TX_8,
	/* 6 */	0,
	/* 7 */	ZSWR7_SDLCFLAG,
	/* 8 */	0,
	/* 9 */	ZSWR9_MASTER_IE | ZSWR9_NO_VECTOR | ZSWR9_VECTOR_INCL_STAT,
	/*10 */	ZSWR10_PRESET_ONES,
	/*11 */	0,
	/*12 */	0,
	/*13 */	0,
	/*14 */	0,
	/*15 */	ZSR15_SDLC_FIFO_ENA | ZSR15_TX_UNDER,
};

/*
 * default `syncmode' (clock and loopback options) for scc.
 */
struct syncmode hsis_default_sm = {
	TXC_IS_BAUD, RXC_IS_BAUD, 1, 0, 9600, 0, 0
};


int hsisidentify(), hsisattach();
struct dev_ops hsis_ops = {
	1,
	hsisidentify,
	hsisattach,
};

int nhsisboard;			/* total # of hsis boards found */

struct hsiscom *hsiscom;
struct hsiscom *hsiscom_end;
struct hsis_softc *hsis_softc;
struct hsis_softc *hsis_softc_end;

int hsis_init(), hsis_output(), hsis_ioctl(), hsis_reset(), hsis_watchdog(),
    hsisintr(), hsissoftint();
void hsis_start();
void rawsyncinit();

void
hsis_scc_attach(unit, hs, addr)
	register u_char unit;
	register struct hsiscom *hs;
	register u_char *addr;
{
	static int nhsischan;
	register struct zscc *zs = &hs->hs_zs[unit];
	register struct zdma *zd = &hs->hs_zd[unit >> 1];
	register u_char *zsdev = addr;
	register struct hsis_softc *sc = &hsis_softc[nhsischan];
	register struct ifnet *ifp = &sc->sc_if;

	zs->zs_unit = unit;
	zs->zs_addr = zsdev;

	/* disable rx, tx and interrupts. */
	SCC_WRITE(1, 0);
	SCC_WRITE(3, 0);
	SCC_WRITE(5, 0);

	/* disable external rcv done interrupt */
	hs->hs_board[unit + HSIS_EINT_ENA_A] = 0;

	/* set up the softc and make us known to the network code. */
	sc->sc_zs = zs;
	sc->sc_zd = zd;
	sc->sc_hs = hs;
	sc->sc_sm = hsis_default_sm;
	sc->sc_fifop = sc->sc_fifo;
	ifp->if_name = "hsis";
	ifp->if_unit = nhsischan++;
	ifp->if_mtu = HSIS_MTU;
#ifdef MULTICAST
	ifp->if_flags = IFF_POINTOPOINT | IFF_MULTICAST;
#else
	ifp->if_flags = IFF_POINTOPOINT;
#endif MULTICAST
	ifp->if_init = hsis_init;
	ifp->if_output = hsis_output;
	ifp->if_ioctl = hsis_ioctl;
	ifp->if_reset = hsis_reset;
	ifp->if_watchdog = hsis_watchdog;
	ifp->if_snd.ifq_maxlen = HSIS_MAX_SENDQ;
#if NBPFILTER > 0
	bpfattach(&sc->sc_bpf, ifp, DLT_PPP, PPP_HDRSPACE);
#endif
	if_attach(ifp);
}

void
hsis_dma_attach(unit, zd, addr)
	register u_char unit;
	register struct zdma *zd;
	register u_char *addr;
{
	register u_char *zddev = addr;

	zd->zd_addr = zddev;

	/*
	 * Reset the dma section. Clear out the `interrupt vector' address
	 * so we get a simple value to switch on when interpreting intr.
	 * Set the DCR to increment addresses rather than decrementing.
	 * Set the ICR to disable intr
	 * on all channels (any event that would generate a dma intr also
	 * generates an scc intr -- we waste enough time dealing with this
	 * stupid chip and don't need to double all the interrupts),
	 * don't put a vector on the bus when req. intr but do include
	 * `status' in the IVR on intr.
	 * Zero out the dma address registers so we don't have to write
	 * the high bytes when switching buffers.
	 */
	DMA_WRITE0(ZSDMA_CCAR_DMA_RESET);
	DMA_WRITE0(ZSDMA_CCAR_RESET_IUS);
	DMA_WRITE(ZSDMA_IVR, 0);
	DMA_WRITE(ZSDMA_ICSR, 0);
	DMA_WRITE(ZSDMA_DCR, 0xf);
	DMA_WRITE(ZSDMA_ICR, ZSDMA_ICR_NV | ZSDMA_ICR_VIS);

	DMA_WRITE(ZSDMA_RDARA, 0);
	DMA_WRITE(ZSDMA_RDARA+1, 0);
	DMA_WRITE(ZSDMA_RDARA+2, 0);
	DMA_WRITE(ZSDMA_RDARA+3, 0);

	DMA_WRITE(ZSDMA_TDARA, 0);
	DMA_WRITE(ZSDMA_TDARA+1, 0);
	DMA_WRITE(ZSDMA_TDARA+2, 0);
	DMA_WRITE(ZSDMA_TDARA+3, 0);

	DMA_WRITE(ZSDMA_RDARB, 0);
	DMA_WRITE(ZSDMA_RDARB+1, 0);
	DMA_WRITE(ZSDMA_RDARB+2, 0);
	DMA_WRITE(ZSDMA_RDARB+3, 0);

	DMA_WRITE(ZSDMA_TDARB, 0);
	DMA_WRITE(ZSDMA_TDARB+1, 0);
	DMA_WRITE(ZSDMA_TDARB+2, 0);
	DMA_WRITE(ZSDMA_TDARB+3, 0);
}

int
hsisidentify(name)
	char *name;
{
	if (strcmp(name, "HSI") == 0) {
		++nhsisboard;
		return (1);
	} else
		return (0);
}

int
hsisattach(dev)
	register struct dev_info *dev;
{
	static int curhsis = 0;
	register struct hsiscom *hs;
	register u_char *hsboard;
	register int i;

	dev->devi_unit = curhsis;
	if (hsiscom == NULL) {
		hsiscom = (struct hsiscom *)new_kmem_zalloc(
				(u_int)(nhsisboard * sizeof (struct hsiscom)),
				KMEM_SLEEP);
		if (hsiscom == NULL) {
			printf("hsis: no space for data structures.\n");
			return (-1);
		}
		hsiscom_end = &hsiscom[nhsisboard];

		hsis_softc = (struct hsis_softc *)new_kmem_zalloc(
			   (u_int)(nhsisboard * 4 * sizeof (struct hsis_softc)),
			   KMEM_SLEEP);
		if (hsis_softc == NULL) {
			printf("hsis: no space for data structures\n");
			return (-1);
		}
		hsis_softc_end = &hsis_softc[nhsisboard * 4];

		rawsyncinit(); /*XXX*/
	}
	hs = &hsiscom[dev->devi_unit];
	hs->hs_dev = dev;

	/*
	 * register our interrupt handler, map the board into kernel memory,
	 * then reset it.  Note that the interrupt level must be <= splimp
	 * (and, given the constraints imposed by the braindead Zilog dma,
	 * the level should be as high as possible, e.g., splimp).
	 */
	addintr(dev->devi_intr->int_pri, hsisintr, dev->devi_name, curhsis);
	addintr(SOFTINT_LEVEL, hsissoftint, "HSI-soft", curhsis);
	hsboard = (u_char *)map_regs(dev->devi_reg->reg_addr,
				     dev->devi_reg->reg_size,
				     dev->devi_reg->reg_bustype);
	hs->hs_board = hsboard;

	/*
	 * set up the free buffer list.
	 */
	for (i = HSIS_SRAM_SIZE; (i -= HSIS_BUFSIZE) >= 0; ) {
		register struct hsisbuf *bp;

		bp = (struct hsisbuf *)(hsboard + i + HSIS_SRAM);
		bp->next = hs->hs_free;
		hs->hs_free = bp;
	}

	/*
	 * Initialize the two dma channels and 4 scc channels.
	 * First reset each chip and set its bus configuration reg.
	 * Then set up the software data structures.
	 */
	hsboard[HSIS_RST_ISSC0] = 0;	/* reset chips */
	hsboard[HSIS_RST_ISSC1] = 0;
	DELAY(10);
	hsboard[HSIS_SCC_A] = 0;	/* clear BCR */
	hsboard[HSIS_SCC_C] = 0;

	hsis_dma_attach(0, &hs->hs_zd[0], &hsboard[HSIS_DMA_0]);
	hsis_dma_attach(1, &hs->hs_zd[1], &hsboard[HSIS_DMA_1]);

	hsis_scc_attach(0, hs, &hsboard[HSIS_SCC_A]);
	hsis_scc_attach(1, hs, &hsboard[HSIS_SCC_B]);
	hsis_scc_attach(2, hs, &hsboard[HSIS_SCC_C]);
	hsis_scc_attach(3, hs, &hsboard[HSIS_SCC_D]);

	report_dev(dev);
	++curhsis;
	return (0);
}

int
copy_m_to_b(sc, m, cp)
	register struct hsis_softc *sc;
	register struct mbuf *m;
	register u_char *cp;
{
	register long len, totlen;
	register struct mbuf *m0 = m;

	totlen = 0;
	do {
		len = m->m_len;
		bcopy(mtod(m, caddr_t), (caddr_t)cp, (u_int)len);
		cp += len;
		totlen += len;
	} while (m = m->m_next);

	m_freem(m0);
	if (sc->sc_raw && (sc->sc_raw->so_snd.sb_flags & SB_WAIT ||
			   sc->sc_raw->so_snd.sb_sel))
		sbwakeup(sc->sc_raw, &sc->sc_raw->so_snd);
	return (totlen);
}

hsis_output(ifp, m, dst)
	register struct ifnet *ifp;
	register struct mbuf *m;
	struct sockaddr *dst;
{
	register int s;
	register struct hsis_softc *sc;

	sc = (struct hsis_softc *)ifp;
	TRACEL(T_OUTPUT, sc, 0)
	if ((ifp->if_flags & IFF_UP) == 0) {
		m_freem(m);
		return (ENETDOWN);
	}
	if (dst->sa_family != AF_UNSPEC) {
		/* Add a PPP header */
		register u_int off = m->m_off;

		if (off < MMINOFF + PPP_HDRSPACE ||
		    (off >= MSIZE && m->m_cltype != MCL_STATIC_HDR)) {
			/* need new mbuf for hdr (should really panic here
			 * then fix whatever isn't leaving space for header) */
			register struct mbuf *m0;

			MGET(m0, M_DONTWAIT, MT_DATA);
			if (m0 == (struct mbuf *)0) {
				m_freem(m);
				return (ENOBUFS);
			}
			m0->m_next = m;
			m0->m_len = PPP_HDRSPACE;
			m = m0;
		} else {
			m->m_off -= PPP_HDRSPACE;
			m->m_len += PPP_HDRSPACE;
		}
#ifdef STII
		if (dst->sa_family == AF_COIP)
		  *mtod(m, u_int *) = PPP_STII;
		else if (dst->sa_family == (0x8000 | AF_COIP))
		  *mtod(m, u_int *) = PPP_SCMP;
		else
#endif STII
		*mtod(m, u_int *) = PPP_INET; /* XXX */
	}
	if (sc->sc_ostate == 0) {
		register struct hsisbuf *bp = sc->sc_curout;
		register int len;

		len = copy_m_to_b(sc, m, BUFtoCP(bp));
		s = splboard();
		TRACE(T_OUT_COPY, sc, (int)bp | len)
		bp->cnt = len;
		sc->sc_ostate = 2;
		hsis_start(sc);
#ifdef HSIS_EXTERNAL_RCVDONE
	} else if (sc->sc_if.if_snd.ifq_len == 0 && sc->sc_nextout->cnt == 0) {
		register struct hsisbuf *bp = sc->sc_nextout;
		register int len;

		len = copy_m_to_b(sc, m, BUFtoCP(bp));
		s = splboard();
		TRACE(T_OUT_COPY, sc, (int)bp | len)
		bp->cnt = len;
		if (sc->sc_ostate == 0) {
			/*
			 * last packet completed while we were doing copy --
			 * flip buffers & restart output.
			 */
			(sc->sc_nextout = sc->sc_curout)->cnt = 0;
			sc->sc_curout = bp;
			sc->sc_ostate = 2;
			hsis_start(sc);
		}
#endif
	} else {
		register struct ifqueue *ifq = &ifp->if_snd;

		s = splboard();
		if (IF_QFULL(ifq)) {
			IF_DROP(ifq);
			TRACE(T_QFULL, sc, 0)
			splx(s);
			m_freem(m);
			return (ENOBUFS);
		}
		IF_ENQUEUE(ifq, m);
		if (sc->sc_ostate == 0)
			hsis_start(sc);
	}
	splx(s);
	return (0);
}

/*
 * start new output operation.  This routine *must* be called at splboard.
 */
void
hsis_start(sc)
	register struct hsis_softc *sc;
{
	register struct zscc *zs = sc->sc_zs;
	register u_char *zsdev = zs->zs_addr;
	register u_char *zddev = sc->sc_zd->zd_addr;
	register u_char unit = sc->sc_if.if_unit;
	register u_char *cp;
	register int len;
	register u_int baddr;
	register int s;
	register u_int resid;
	register int i;

	cp = BUFtoCP(sc->sc_curout);
	if (sc->sc_ostate == 0) {
		register struct hsisbuf *bp;

		if (len = (bp = sc->sc_nextout)->cnt) {
			/*
			 * `next' output buffer is full - swap current and
			 * next.  (The weird assignment below fools Sun-4 cc
			 * into generating reasonable code -- maybe one day
			 * Sun will discover ANSI C `volatile' and this
			 * crap can go away.)
			 */
			(sc->sc_nextout = sc->sc_curout)->cnt = 0;
			sc->sc_curout = bp;
			cp = BUFtoCP(bp);
		} else {
			/*
			 * on-board xmit buffer is empty.  try to copy a new
			 * packet to it.  (we want to overlap the copy with
			 * the scc sending the final crc and flag bytes to
			 * avoid taking an extra xmit interrupt).
			 */
			register struct mbuf *m;

			IF_DEQUEUE(&sc->sc_if.if_snd, m);
			if (m == NULL) {
				sc->sc_if.if_timer = 0;
				TRACE(T_EMPTY_OQ, sc, 0)
				return;
			}
			len = copy_m_to_b(sc, m, cp);
			TRACE(T_OUT_COPY, sc, (int)bp | len)
		}
	} else
		len = sc->sc_curout->cnt;

	/*
	 * if the transmit buffer is full it means that crc/flag
	 * sending is still in progress.  the stupid scc will jam
	 * two packets together if we enable dma xmit so we have
	 * to turn on the xmit interrupt (which should come when
	 * the crc and flag have been sent) and exit waiting for
	 * that interrupt.  If the transmit buffer is empty, we
	 * can just start the next packet (this should be the
	 * usual case at T1 speeds).
	 */
	sc->sc_if.if_timer = HSIS_WATCHDOG_TIME;
	SCC_READ0(resid);
	if ((resid & ZSRR0_TX_READY) == 0) {
		/*
		 * we lose - wait for xmit intr.
		 */
		SCC_BIS(1, ZSWR1_TIE);
		sc->sc_ostate = 2;
		sc->sc_curout->cnt = len;
		return;
	}
	/*
	 * Transmit buffer empty -- start up dma.
	 */
	baddr = cp - sc->sc_hs->hs_board;
	SCC_WRITE0(ZSWR0_RESET_TXCRC);
	if (unit & 1) {
		/* B channel */
		register u_int c;

		DMA_READ(ZSDMA_TDCRB, resid);
		DMA_READ(ZSDMA_TDCRB + 1, c);
		resid |= c << 8;

		DMA_WRITE(ZSDMA_TDCRB, len);
		DMA_WRITE(ZSDMA_TDCRB+1, len >> 8);

		DMA_WRITE(ZSDMA_TDARB, baddr);
		DMA_WRITE(ZSDMA_TDARB+1, baddr >> 8);

		s = splhigh();
		DMA_WRITE(ZSDMA_CCAR, ZSDMA_CCAR_ENA_TX_B);
	} else {
		/* A channel */
		register u_int c;

		DMA_READ(ZSDMA_TDCRA, resid);
		DMA_READ(ZSDMA_TDCRA + 1, c);
		resid |= c << 8;

		DMA_WRITE(ZSDMA_TDCRA, len);
		DMA_WRITE(ZSDMA_TDCRA+1, len >> 8);

		DMA_WRITE(ZSDMA_TDARA, baddr);
		DMA_WRITE(ZSDMA_TDARA+1, baddr >> 8);

		s = splhigh();
		DMA_WRITE(ZSDMA_CCAR, ZSDMA_CCAR_ENA_TX_A);
	}
	/*
	 * Cretinous chip requires that EOM not be reset until dma has
	 * loaded first data character into buffer but must be reset
	 * before last character loaded into buffer.  We've locked out
	 * to prevent the obvious race so just wait until character gets
	 * there.
	 */
	for (i = 10; --i >= 0; ) {
		DELAY(2);
		SCC_READ0(baddr);
		if ((baddr & ZSRR0_TX_READY) == 0)
			goto rdy;
	}
	TRACE(T_NO_XMIT_RDY, sc, baddr)
	dprintf(("hsis%d: hsis_start: xmit didn't load (rr0=0x%x)\n",
		unit, baddr))
rdy:
	SCC_WRITE0(ZSWR0_RESET_EOM);
	TRACE(T_START_OUT, sc, (resid << 16) | len)
	sc->sc_ostate = 1;
	splx(s);
	if (resid) {
		++sc->sc_if.if_oerrors;
		++sc->sc_estats.sse_underrun;
	}
	++sc->sc_if.if_opackets;
	++sc->sc_dstats.ssd_opack;
	sc->sc_dstats.ssd_ochar += len;
#if NBPFILTER > 0
	if (sc->sc_bpf)
		bpf_tap(sc->sc_bpf, cp, len);
#endif
#ifdef HSIS_EXTERNAL_RCVDONE
	/*
	 * if there's more in the snd q, copy another packet to the
	 * on-board `next' buffer (we do it now to overlap the copy
	 * with the send of the last packet to approximate back-to-back
	 * output packets).
	 */
	if (sc->sc_if.if_snd.ifq_len) {
		register struct mbuf *m;

		IF_DEQUEUE(&sc->sc_if.if_snd, m);
		if (m) {
			register struct hsisbuf *bp = sc->sc_nextout;

			bp->cnt = len = copy_m_to_b(sc, m, BUFtoCP(bp));
			TRACE(T_OUT_COPY, sc, (int)bp | len)
		}
	}
#endif
}

int
hsis_start_dma_read(sc)
	register struct hsis_softc *sc;
{
	register u_char *zddev = sc->sc_zd->zd_addr;
	register int baddr = BUFtoCP(sc->sc_inbuf) - sc->sc_hs->hs_board;
	register u_char bl = baddr;
	register u_char bh = baddr >> 8;
	register int s = splhigh();
	register u_int rl;
	register u_char rh;

	/*
	 * disable the dma channel, set it to xfer into sc_inbuf,
	 * then re-enable.  If there are back-to-back packets inbound,
	 * we have to re-enable the dma channel before the 3 byte scc
	 * rcv fifo overflows.  E.g., at T1 (5.2us/byte) the time
	 * from disable to enable should be no more than 10us.  So,
	 * we make sure nothing interrupts us during this window and
	 * the code below is as fast as I can make it.
	 * Data from the next packet probably got stuffed into the
	 * buffer containing the current packet so we save and return
	 * the current dma count reg. before overwriting it (so the
	 * higher level routine can undo the damage done by this
	 * braindead dma model).
	 */
	/* spin until the rcv fifo is empty */
	for (rl = 10; --rl != 0; ) {
		register u_char *zsdev = sc->sc_zs->zs_addr;
		SCC_READ0(rh)
		if ((rh & ZSRR0_RX_READY) == 0)
			break;
		DELAY(2);
	}
	if (rl == 0)
		TRACE(T_FAIL, sc, rh)

	if (sc->sc_zs->zs_unit & 1) {
		/* B channel */
		DMA_BIC(ZSDMA_DER, ZSDMA_DER_RX_B_ENABLE);

		DMA_WRITE(ZSDMA_RDCRB, HSIS_MAXPACKET & 0xff);
		DMA_WRITE(ZSDMA_RDCRB+1, HSIS_MAXPACKET >> 8);

		DMA_READ(ZSDMA_RDARB, rl);
		DMA_WRITE(ZSDMA_RDARB, bl);
		DMA_READ(ZSDMA_RDARB+1, rh);
		DMA_WRITE(ZSDMA_RDARB+1, bh);

		DMA_WRITE0(ZSDMA_CCAR_ENA_RX_B);
	} else {
		/* A channel */
		DMA_BIC(ZSDMA_DER, ZSDMA_DER_RX_A_ENABLE);

		DMA_WRITE(ZSDMA_RDCRA, HSIS_MAXPACKET & 0xff);
		DMA_WRITE(ZSDMA_RDCRA+1, HSIS_MAXPACKET >> 8);

		DMA_READ(ZSDMA_RDARA, rl);
		DMA_WRITE(ZSDMA_RDARA, bl);
		DMA_READ(ZSDMA_RDARA+1, rh);
		DMA_WRITE(ZSDMA_RDARA+1, bh);

		DMA_WRITE0(ZSDMA_CCAR_ENA_RX_A);
	}
	rl |= rh << 8;
	TRACE(T_START_READ, sc, (((bh << 8) | bl) << 16) | rl)
	splx(s);
	return ((int)(rl - sizeof(struct hsisbuf)) & (HSIS_BUFSIZE - 1));
}

/*
 * This routine is called if an output operation takes longer than
 * HSIS_WATCHDOG_TIME (usually 2 minutes) to complete.  Reset and
 * restart the channel (current output packet will be lost).
 */
hsis_watchdog(unit)
	register int unit;
{
	register struct hsis_softc *sc = &hsis_softc[unit];
	register u_int errcnt = sc->sc_oerrcnt;
	register int s;

	if (sc->sc_inbuf == NULL || (sc->sc_if.if_flags & IFF_UP))
		/* we have been manually turned offline or online */
		return;

	if (errcnt == 0)
		printf("hsis%d: watchdog timeout.\n", unit);
	s = splboard();
	sc->sc_oerrcnt = ++errcnt;
	hsis_reset(unit);
	hsis_init(unit);
	hsis_start(sc);
	splx(s);
}

void
hsis_ierr_timer(sc)
	register struct hsis_softc *sc;
{
	register int unit = sc->sc_if.if_unit;
	register int s;

	if (sc->sc_inbuf == NULL || (sc->sc_if.if_flags & IFF_UP))
		/* we have been manually turned offline or online */
		return;
	printf("hsis%d: reset and restarted.\n", unit);
	s = splboard();
	hsis_init(unit);
	hsis_start(sc);
	splx(s);
}

void
hsis_ierror(sc, zs, zsdev, msg)
	register struct hsis_softc *sc;
	register struct zscc *zs;
	register u_char *zsdev;
	register char *msg;
{
	register int unit = sc->sc_if.if_unit;
	register u_int errcnt = sc->sc_ierrcnt;

	sc->sc_ierrcnt = ++errcnt;
	if (errcnt >= HSIS_RESET_THRESH) {
		hsis_reset(unit);
		if (errcnt >= HSIS_OFF_THRESH)
			timeout(hsis_ierr_timer, sc, HSIS_OFF_TIME);
		else
			hsis_ierr_timer(sc);
		return;
	}
	if (errcnt == 1)
		printf("hsis%d: %s (p#%d).\n", unit, msg,
			sc->sc_dstats.ssd_ipack);

	/*
	 * book says we have to disable then re-enable
	 * fifo to clear it.  Then toss input queue in (vain) hope that
	 * we'll end up with input stream and fifo in sync.
	 */
*(u_char *)AUXIO_REG = AUX_MBO|AUX_EJECT;
	SCC_BIC(3, ZSWR3_RX_ENABLE)
	SCC_BIC(15, ZSR15_SDLC_FIFO_ENA)
	++sc->sc_if.if_ierrors;
	SCC_BIS(15, ZSR15_SDLC_FIFO_ENA)
	SCC_WRITE0(ZSWR0_RESET_ERRORS);
	(void) hsis_start_dma_read(sc);
	SCC_BIS(3, ZSWR3_RX_ENABLE)
	if (sc->sc_intail) {
		sc->sc_intail->next = sc->sc_hs->hs_free;
		sc->sc_hs->hs_free = sc->sc_inq;
		sc->sc_inq = NULL;
		sc->sc_intail = NULL;
	}
	bzero(sc->sc_fifo, sizeof(sc->sc_fifo));
	sc->sc_fifop = sc->sc_fifo;
	sc->sc_inoff = NULL;
*(u_char *)AUXIO_REG = AUX_MBO|AUX_EJECT|AUX_LED;
}

void
hsis_sccrecv_intr(sc, zs, zsdev)
	register struct hsis_softc *sc;
	register struct zscc *zs;
	register u_char *zsdev;
{
	register struct hsiscom *hs;
	register struct hsisbuf *bp;
	register struct hsisbuf *qp;
	register int *fp;

	/*
	 * Stash the current contents of the sdlc fifo.  This should
	 * be done before we call `start_dma_read' or we can end up
	 * with the fifo & dma counts different:  For god-only-knows
	 * what reason, Zilog made the sdlc fifo count bytes that have
	 * entered the 3 byte receive data fifo while the dma counts
	 * bytes that have left the receive data fifo.  Thus the fifo
	 * can record the end of a packet that the dma hasn't finished.
	 */
	fp = sc->sc_fifop;
	while (1) {
		register u_char rr1, rr6, rr7;
		register int i;

		SCC_READ(7, rr7)
		SCC_READ(6, rr6)
		SCC_READ(1, rr1)
		i = (((rr1 << 8) | rr7) << 8) | rr6;
		TRACE(T_RCVINT_FIFO, sc, i)
		if (rr1 & ZSRR1_DO) {
			/*
			 * data overrun - the way the Zilog fifo works
			 * we don't have a prayer of recovering so toss
			 * everything in the fifo and input queue, put
			 * the scc in hunt mode to skip to the start of
			 * the next packet, then hope the fifo eventually
			 * gets back in sync with the input stream.
			 */
			hsis_ierror(sc, zs, zsdev, "receive data overrun");
			++sc->sc_estats.sse_overrun;
			return;
		}
		switch (rr7 & 0xc0) {

		case 0x80:
		case 0xc0:
			/* fifo overflow */
			hsis_ierror(sc, zs, zsdev, "status fifo overflow");
			return;

		case 0x00:
			/* fifo empty */
			goto fifo_empty;
		}
		if (*fp) {
			hsis_ierror(sc, zs, zsdev, "status fifo array overflow");
			return;
		}
		*fp++ = i;
		if (fp >= &sc->sc_fifo[HSIS_NFIFO])
			fp = sc->sc_fifo;
	}
fifo_empty:
	if (fp == sc->sc_fifop)
		/* did nothing (spurious read interrupt) */
		return;
	sc->sc_fifop = fp;

	hs = sc->sc_hs;
	if ((bp = sc->sc_inbuf) == 0) {
		/* channel is offline */
		(void)hsis_start_dma_read(sc);
		return;
	}
	if ((qp = hs->hs_free) == NULL) {
		/* no buffer - have to toss input or we lose sync with fifo */
		TRACE(T_RCVINT_BUF, sc, 0)
		hsis_ierror(sc, zs, zsdev, "no free bufs");
		return;
	}
	hs->hs_free = qp->next;
	sc->sc_inbuf = qp;
	if ((bp->cnt = hsis_start_dma_read(sc)) == 0) {
		bp->next = hs->hs_free;
		hs->hs_free = bp;
	} else {
		bp->next = NULL;
		if (qp = sc->sc_intail)
			qp->next = bp;
		else
			sc->sc_inq = bp;
		sc->sc_intail = bp;
	}
	SOFTINT(sc, RECV_DONE);
}

void
hsis_stat_intr(sc, zs, zsdev)
	register struct hsis_softc *sc;
	register struct zscc *zs;
	register u_char *zsdev;
{
	register u_int rr0;

	SCC_READ0(rr0);
	SCC_WRITE0(ZSWR0_RESET_STATUS);
	SCC_WRITE0(ZSWR0_CLR_INTR);
	TRACE(T_STAT_INT, sc, rr0)
	if ((rr0 & ZSRR0_TXUNDER) && sc->sc_ostate) {
		/* packet completed - start a new one if possible */
		sc->sc_oerrcnt = 0;
		sc->sc_ostate = 0;
		SOFTINT(sc, XMIT_DONE);
	} else if (rr0 & ZSRR0_BREAK) {
		/*
		 * `abort' received -- reset input (because scc fifo
		 * state is messed up and we have no way to figure
		 * out packet boundaries).
		 */
		hsis_ierror(sc, zs, zsdev, "received `abort'");
		++sc->sc_estats.sse_abort;
	} else {
		/* XXX - should do something. */
		dprintf(("hsis%d: scc stat interrupt, rr0=0x%x.\n",
			sc->sc_if.if_unit, rr0))
	}
}

void
hsis_sccxmit_intr(sc, zs, zsdev)
	register struct hsis_softc *sc;
	register struct zscc *zs;
	register u_char *zsdev;
{
	register u_int rr0;

	SCC_READ0(rr0);
	SCC_BIC(1, ZSWR1_TIE);
	SCC_WRITE0(ZSWR0_RESET_TXINT);
	SCC_WRITE0(ZSWR0_CLR_INTR);
	TRACE(T_XMIT_INT, sc, (sc->sc_ostate << 16) | rr0)
	if (sc->sc_ostate == 2) {
		/*
		 * we were waiting for CRC/flag send to finish and it
		 * has -- start next packet.
		 */
		SOFTINT(sc, XMIT_DONE);
	} else {
		/*
		 * since we never enabled xmit interrupt, something's weird.
		 * it would be nice to print a warning message but there's
		 * a small race in hsis_start where we TIE then find xmit is
		 * done & turn it off.  Zilog says the disable should clear
		 * the interrupt but, of course, they lie & we can end
		 * up here.
		 */
	}
}

#define ZS_A_INTR (ZSRR3_IP_A_STAT|ZSRR3_IP_A_RX|ZSRR3_IP_A_TX)
#define ZS_B_INTR (ZSRR3_IP_B_STAT|ZSRR3_IP_B_RX|ZSRR3_IP_B_TX)

#define ZINTSCAN(rmask, smask, tmask) { \
	if (rr3 & rmask) { \
		SCC_WRITE0(ZSWR0_RESET_ERRORS); \
		SCC_WRITE0(ZSWR0_CLR_INTR); \
		hsis_sccrecv_intr(sc, zs, zsdev); \
	} \
	if (rr3 & smask) \
		hsis_stat_intr(sc, zs, zsdev); \
	if (rr3 & tmask) \
		hsis_sccxmit_intr(sc, zs, zsdev); \
}

#define	ZCHECK_INT(chan) ZINTSCAN(ZSRR3_IP_/**/chan/**/_RX, \
				  ZSRR3_IP_/**/chan/**/_STAT, \
				  ZSRR3_IP_/**/chan/**/_TX)
#ifndef HSIS_EXTERNAL_RCVDONE
#define ZCHECK_BOTH { \
	zs = sc->sc_zs; \
	zsdev = zs->zs_addr; \
	SCC_READ(3, rr3); \
	if (rr3 & ZS_A_INTR) { \
		++did_something; \
		ZCHECK_INT(A) \
	} \
	++sc; \
	if (rr3 & ZS_B_INTR) { \
		++did_something; \
		zs = sc->sc_zs; \
		zsdev = zs->zs_addr; \
		ZCHECK_INT(B) \
	} \
	++sc; \
}
#else
#define ZCHECK_BOTH { \
	zs = sc->sc_zs; \
	zsdev = zs->zs_addr; \
	if ((hsis[HSIS_EINT_STS] & rcvintmask) == 0) {\
		++did_something; \
		hsis[zs->zs_unit + HSIS_EINT_CLR_A] = 0; \
		hsis_sccrecv_intr(sc, zs, zsdev); \
	} \
	rcvintmask <<= 1; \
	SCC_READ(3, rr3); \
	if (rr3 & ZS_A_INTR) { \
		++did_something; \
		ZCHECK_INT(A) \
	} \
	++sc; \
	if ((hsis[HSIS_EINT_STS] & rcvintmask) == 0) {\
		++did_something; \
		zs = sc->sc_zs; \
		zsdev = zs->zs_addr; \
		hsis[zs->zs_unit + HSIS_EINT_CLR_A] = 0; \
		hsis_sccrecv_intr(sc, zs, zsdev); \
	} \
	rcvintmask <<= 1; \
	if (rr3 & ZS_B_INTR) { \
		++did_something; \
		zs = sc->sc_zs; \
		zsdev = zs->zs_addr; \
		ZCHECK_INT(B) \
	} \
	++sc; \
}
#endif

/*
 * Handle HSIS board(s) interrupt(s).
 */
int
hsisintr()
{
	register struct hsis_softc *sc;
	register struct hsis_softc *sc_end = hsis_softc_end;
	register struct zscc *zs;
	register u_char *zsdev, *hsis;
	register u_char rr3;
	register int did_something = 0, last_round;

	/*
	 * It's costly to take an interrupt and likely that some other
	 * channel finished while we were servicing the current channel.
	 * We loop here until we make one full pass through the status
	 * registers and don't find new work to do.
	 */
	TRACE(T_INTR_ENTRY, hsis_softc, 0)
	do {
		last_round = did_something;
		/*
		 * we can't reliably read the status registers on the
		 * zilog chip so we look at the hsis board status register
		 * to find interrupt requests.  This means we loop over
		 * four zilog channels at a time.
		 */
		for (sc = hsis_softc; sc < sc_end; ) {
#ifdef HSIS_EXTERNAL_RCVDONE
			register u_char rcvintmask = 1;
#endif
			hsis = sc->sc_hs->hs_board;
			ZCHECK_BOTH
			ZCHECK_BOTH
		}
	} while (did_something != last_round);

	if (hsis_isum)
		set_intreg(IR_SOFT_INT4, 1);

	TRACE(T_INTR_EXIT, hsis_softc, did_something)
	return (did_something);
}

void
hsis_drop_input(sc, len, bufoff, buflen)
	register struct hsis_softc *sc;
	register int len, buflen;
	register u_char *bufoff;
{
	/*
	 * Note:  To avoid races between the hardware & software
	 * interrupt levels, it's important that the load of 'bp'
	 * below be done *after* we are at splboard.  In particular,
	 * this means that the load of bp *cannot* be put in the
	 * delay slot of the spl call.  The statement order below
	 * works with Sun's current compiler technology but may give
	 * problems in the future (one day they'll discover "volatile"...)
	 */
	register int s = splboard();
	register struct hsiscom *hs = sc->sc_hs;
	register struct hsisbuf *bp = sc->sc_inq;

	if (bp == NULL) {
		/*
		 * timeout or hardware intr probably did a reset --
		 * exit, making sure that input state stays clean.
		 */
		sc->sc_inoff = 0;
		splx(s);
		return;
	}
	while (len) {
		if (buflen <= 0) {
			buflen = bp->cnt;
			bufoff = BUFtoCP(bp);
		}
		if (buflen > len) {
			buflen -= len;
			bufoff += len;
			break;
		} else {
			register struct hsisbuf *np;

			len -= buflen;
			buflen = 0;
			bufoff = NULL;

			if ((sc->sc_inq = np = bp->next) == NULL)
				sc->sc_intail = NULL;
			bp->next = hs->hs_free;
			hs->hs_free = bp;
			bp = np;
		}
	}
	if (buflen)
		bp->cnt = buflen;
	sc->sc_inoff = bufoff;
	splx(s);
}

/*
 * Following is mbuf offset to get nice alignment of packets.
 * Choice determined by:
 *  1. IP header *must* be aligned on a word (4 byte) boundary.
 *  2. bcopy will go much faster if data is cache aligned (i.e., at
 *     16-byte boundary).
 *  3. We need space to prepend a 14 byte ethernet header (if
 *     forwarding packet).
 *
 * MMINOFF is 12 (thanks to Bill Joy for this horrible kludge) and the
 * next mult-of-16 above 12+14 is 32 so we offset 20 (= 32 - MMINOFF)
 * in an mbuf and 16 in a cluster.
 */
#define HDRSPACE (16)
#define MBUF_HDRSPACE (32 - MMINOFF)

/*
 * following macro drops `len' input bytes (used on input errors).
 */
#define DROP_INPUT(len) { \
	hsis_drop_input(sc, len, bufoff, buflen); \
	if ((bp = sc->sc_inq) == NULL) \
		/* input was reset by timeout or error */ \
		return; \
	if (bufoff = sc->sc_inoff) \
		buflen = bp->cnt; \
	else \
		buflen = 0; \
}

/*
 * This routine goes through the software copy of the scc
 * status fifo (which records the lengths of incoming packets)
 * and sc_inq (the queue of unprocessed, incoming data) and
 * breaks the data up into packets then queues them for higher
 * level network processing.  The loop structure is complicated
 * by the fact that packet boundaries are mapped randomly onto
 * the input queue (there may be more than one packet per input
 * buffer and packets may cross buffer boundaries).
 */
void
hsis_recv_done(sc)
	register struct hsis_softc *sc;
{
	register int *fp;
	register u_char *bufoff;
	register int buflen;
	register struct hsiscom *hs = sc->sc_hs;
	register struct hsisbuf *bp;

	/* find the first unprocessed packet in the fifo array */
	for (fp = sc->sc_fifop - 1; ; --fp) {
		if (fp < sc->sc_fifo)
			fp = &sc->sc_fifo[HSIS_NFIFO - 1];
		if (*fp == 0)
			break;
	}

	if (bufoff = sc->sc_inoff) {
		if ((bp = sc->sc_inq) == NULL)
			return;
		buflen = bp->cnt;
	} else
		buflen = 0;

	while (1) {
		register struct mbuf *m;
		register u_char *op;
		register int sfifo;

		if (++fp >= &sc->sc_fifo[HSIS_NFIFO])
			fp = sc->sc_fifo;

		if ((sfifo = *fp) == 0)
			break;
		*fp = 0;
		TRACEL(T_RCVINT, sc, sfifo)

		if (sfifo & (ZSRR1_FE << 16)) {
			++sc->sc_estats.sse_crc;
			++sc->sc_if.if_ierrors;
			sfifo &= 0x3fff;
			DROP_INPUT(sfifo)
			continue;
		}
		sfifo &= 0x3fff;
		if (sfifo > MCLBYTES - HDRSPACE) {
			/* packet too big */
			++sc->sc_if.if_ierrors;
			DROP_INPUT(sfifo)
			continue;
		}
		MGET(m, M_DONTWAIT, MT_DATA);
		if (m == (struct mbuf *)0) {
			DROP_INPUT(sfifo)
			continue;
		}
		if (sfifo <= MLEN - MBUF_HDRSPACE)
			m->m_off += MBUF_HDRSPACE;
		else {
			/* too big for mbuf - use cluster */
			MCLGET(m);
			if (m->m_len != MCLBYTES) {
				/* no clusters - drop this packet */
				m_freem(m);
				DROP_INPUT(sfifo)
				continue;
			}
			m->m_off += HDRSPACE;
		}
		if (! sc->sc_raw)
			m->m_off -= PPP_HDRSPACE;
		m->m_len = sfifo - 2;
		sc->sc_dstats.ssd_ichar += sfifo;
		++sc->sc_dstats.ssd_ipack;
		++sc->sc_if.if_ipackets;
		op = mtod(m, u_char *);
		while (sfifo) {
			if (buflen <= 0) {
				bp = sc->sc_inq;
				if (bp == NULL) {
					/*
					 * nothing on input queue - probably
					 * had an overrun at hardware intr
					 * level.  just bail.
					 */
					m_freem(m);
					return;
				}
				buflen = bp->cnt;
				bufoff = BUFtoCP(bp);
			}
			if (buflen > sfifo) {
				bcopy(bufoff, op, sfifo);
				buflen -= sfifo;
				bufoff += sfifo;
				break;
			} else {
				register int s;
				register struct hsisbuf *bp;

				bcopy(bufoff, op, buflen);
				sfifo -= buflen;
				op += buflen;
				buflen = 0;
				bufoff = NULL;

				s = splboard();
				if ((bp = sc->sc_inq) == NULL) {
					splx(s);
					m_freem(m);
					return;
				}
				if ((sc->sc_inq = bp->next) == NULL)
					sc->sc_intail = NULL;
				bp->next = hs->hs_free;
				hs->hs_free = bp;
				splx(s);
			}
		}
		/*
		 * queue the packet to the appropriate network protocol.
		 */
#if NBPFILTER > 0
		if (sc->sc_bpf) {
			register u_char c;

			op = mtod(m, u_char *);
			c = *op;
			*op = 0;	/* `in' direction */
			bpf_tap(sc->sc_bpf, op, m->m_len);
			*op = c;
		}
#endif
		if (sc->sc_raw) {
			register struct socket *so = sc->sc_raw;

			if (m->m_len > sbspace(&so->so_rcv)) {
				++sc->sc_idrops;
				m_freem(m);
			} else {
				register struct mbuf *n = so->so_rcv.sb_mb;

				if (n) {
					while (n->m_next)
						n = n->m_next;
					n->m_next = m;
				} else
					so->so_rcv.sb_mb = m;

				so->so_rcv.sb_cc += m->m_len;
				sorwakeup(so);
			}
		} else {
			register struct ifqueue *inq;
			register int s;
#ifdef STII
			struct ifqueue *st2_fromhsis();
			int st2_intr();
#endif STII

			/*
			 * Note: the protocol input routines all require
			 * an ifp at the front of the buffer.  We make
			 * use of the fact that a PPP header is the same
			 * size as an ifp & just overwrite the 4 bytes
			 * of header without allocating new space.
			 */
			s = splimp();
			op = mtod(m, u_char *);
			switch (*(u_int *)op) {

			case PPP_INET:
				inq = &ipintrq;
				schednetisr(NETISR_IP);
				break;
#ifdef STII
			case PPP_SCMP:
			case PPP_STII:
				inq = st2_fromhsis( &sc->sc_if, m, op );
				if ( ! inq ) {
					splx(s);
					sc->sc_ierrcnt = 0;
					continue;
				}
				softcall(st2_intr, (caddr_t)0);
				break;
#endif STII
			default:
				splx(s);
				++sc->sc_ibadtype;
				m_freem(m);
				continue;
			}
			if (IF_QFULL(inq)) {
				++sc->sc_idrops;
				m_freem(m);
			} else {
				*(struct ifnet **)op = &sc->sc_if;
				IF_ENQUEUE(inq, m);
			}
			splx(s);
		}
		sc->sc_ierrcnt = 0;
	}
	/*
	 * done with fifo -- if there's data left in the current buffer,
	 * remember where we are.
	 */
	if (bufoff && (bp = sc->sc_inq))
		bp->cnt = buflen;
	sc->sc_inoff = bufoff;
}

int
hsissoftint()
{
	register u_int isum;
	register int did_something = 0;

	TRACE(T_SOFTINT, hsis_softc, hsis_isum)
	while (1) {
		register struct hsis_softc *sc;
		register struct hsis_softc *sc_end = hsis_softc_end;
		register int s;

		s = splboard();
		isum = hsis_isum;
		hsis_isum = 0;
		splx(s);
		if (isum == 0)
			break;

		++did_something;
		sc = hsis_softc;
		for ( ; isum && sc < sc_end; ++sc) {
			if (isum & XMIT_DONE) {
				s = splboard();
				hsis_start(sc);
				splx(s);
			}
			if (isum & RECV_DONE)
				hsis_recv_done(sc);
			isum >>= 2;
		}
	}
	TRACE(T_SOFTINT_EXIT, hsis_softc, did_something)
	return (did_something);
}

int
hsis_ioctl(ifp, cmd, data)
	register struct ifnet *ifp;
	register int cmd;
	register caddr_t data;
{
	register int unit = ifp->if_unit;
	register struct hsis_softc *sc = &hsis_softc[unit];
	register struct ifreq *ifr = (struct ifreq *)data;
	register int s = splboard(), error = 0;

	TRACE(T_IOCTL, sc, cmd)
	switch (cmd) {

	case SIOCSIFADDR:
		bzero((caddr_t)&sc->sc_dstats, sizeof(sc->sc_dstats));
		bzero((caddr_t)&sc->sc_estats, sizeof(sc->sc_estats));
		error = hsis_init(unit);
		break;

	case SIOCSIFDSTADDR:
		break;

	case SIOCSIFFLAGS:
		switch (ifp->if_flags & (IFF_UP|IFF_RUNNING)) {

		case IFF_UP:
			/* down interface just marked up */
			bzero((caddr_t)&sc->sc_dstats, sizeof(sc->sc_dstats));
			bzero((caddr_t)&sc->sc_estats, sizeof(sc->sc_estats));
			error = hsis_init(unit);
			hsis_start(sc);
			break;

		case IFF_RUNNING:
			/* up interface just marked down */
			hsis_offline(unit);
			break;
		}
		break;

	case SIOCSSDSTATS:
		*(struct ss_dstats *)ifr->ifr_data = sc->sc_dstats;
		break;

	case SIOCSSESTATS:
		*(struct ss_estats *)ifr->ifr_data = sc->sc_estats;
		break;

	case SIOCGETSYNC:
		*(struct syncmode *)ifr->ifr_data = sc->sc_sm;
		break;

	case SIOCSETSYNC:
		if (ifp->if_flags & IFF_RUNNING)
			hsis_offline(unit);
		sc->sc_sm = *(struct syncmode *)ifr->ifr_data;
		if (sc->sc_sm.sm_baudrate != 0) {
			bzero((caddr_t)&sc->sc_dstats, sizeof(sc->sc_dstats));
			bzero((caddr_t)&sc->sc_estats, sizeof(sc->sc_estats));
			error = hsis_init(unit);
			hsis_start(sc);
		}
		break;

#ifdef MULTICAST
	case SIOCADDMULTI:
	case SIOCDELMULTI:
		switch (ifr->ifr_addr.sa_family) {
#ifdef INET
		case AF_INET:
			break;
#endif INET
		default:
			error = EAFNOSUPPORT;
			break;
		}
		break;
#endif MULTICAST

	default:
		error = EINVAL;
	}
	splx(s);
	return (error);
}

int
hsis_init(unit)
	int unit;
{
	register struct hsis_softc *sc = &hsis_softc[unit];
	register struct zscc *zs = sc->sc_zs;
	register u_char *zsdev = zs->zs_addr;
	u_char wreg[sizeof(zs->zs_wreg)];
	register int s;
	register int clk = 0, txin = 0;

	TRACEL(T_INIT, sc, sc->sc_if.if_flags)
	if (sc->sc_if.if_flags & IFF_RUNNING)
		return (0);

	/* set appropriate defaults for scc */
	bcopy((caddr_t)hsis_scc_setup, (caddr_t)wreg, sizeof(wreg));

	/*
	 * modify defaults as per Sun's `syncmode' (hsi compatibility).
	 */
	switch (sc->sc_sm.sm_txclock) {

	case TXC_IS_TXC:
		wreg[11] |= ZSWR11_TXCLK_TRXC;
		txin = 1;
		break;

	case TXC_IS_RXC:
		wreg[11] |= ZSWR11_TXCLK_RTXC;
		break;

	case TXC_IS_BAUD:
		wreg[11] |= ZSWR11_TXCLK_BAUD;
		clk = 1;
		break;

	case TXC_IS_PLL:
		wreg[11] |= ZSWR11_TXCLK_DPLL;
		clk = 2;
		break;

	default:
		return (EINVAL);
	}
	switch (sc->sc_sm.sm_rxclock) {

	case RXC_IS_RXC:
		wreg[11] |= ZSWR11_RXCLK_RTXC;
		break;

	case RXC_IS_TXC:
		wreg[11] |= ZSWR11_RXCLK_TRXC;
		txin = 1;
		break;

	case RXC_IS_BAUD:
		wreg[11] |= ZSWR11_RXCLK_BAUD;
		clk |= 1;
		break;

	case RXC_IS_PLL:
		wreg[11] |= ZSWR11_RXCLK_DPLL;
		clk |= 2;
		break;

	default:
		return (EINVAL);
	}
	if (clk) {
		register long tconst;

		if (clk > 2)
			return (EINVAL);
		tconst = sc->sc_sm.sm_baudrate;
		if (clk == 2) {
			if (! sc->sc_sm.sm_nrzi)
				return (EINVAL);
			tconst <<= 5;
		}
		tconst = HSIS_PCLK / (tconst * 2) - 2;
		if (tconst == 0)
			return (EINVAL);

		sc->sc_sm.sm_baudrate = (HSIS_PCLK / 2) / (tconst + 2);

		wreg[12] = tconst;
		wreg[13] = tconst >> 8;
		wreg[14] |= ZSWR14_BAUD_FROM_PCLK | ZSWR14_BAUD_ENA;
	}
	if (! txin)
		wreg[11] |= ZSWR11_TRXC_OUT_ENA | ZSWR11_TRXC_XMIT;
	if (sc->sc_sm.sm_loopback)
		wreg[14] |= ZSWR14_LOCAL_LOOPBACK;
	if (sc->sc_sm.sm_nrzi)
		wreg[10] |= ZSWR10_NRZI;

	/*
	 * if we don't have read and write buffers yet, get them.
	 */
	s = splboard();
	if (sc->sc_inbuf == NULL) {
		if ((sc->sc_inbuf = sc->sc_hs->hs_free) == NULL) {
			printf("hsis%d: hsis_init: no free bufs.\n",
				sc->sc_if.if_unit);
			splx(s);
			return (ENOBUFS);
		}
		sc->sc_hs->hs_free = sc->sc_inbuf->next;
	}
	if (sc->sc_curout == NULL) {
		if ((sc->sc_curout = sc->sc_hs->hs_free) == NULL) {
			printf("hsis%d: hsis_init: no free bufs\n",
				sc->sc_if.if_unit);
			splx(s);
			return (ENOBUFS);
		}
		sc->sc_hs->hs_free = sc->sc_curout->next;
		sc->sc_curout->cnt = 0;
	}
	if (sc->sc_nextout == NULL) {
		if ((sc->sc_nextout = sc->sc_hs->hs_free) == NULL) {
			printf("hsis%d: hsis_init: no free bufs\n",
				sc->sc_if.if_unit);
			splx(s);
			return (ENOBUFS);
		}
		sc->sc_hs->hs_free = sc->sc_nextout->next;
		sc->sc_nextout->cnt = 0;
	}

	/*
	 * If we got here, sm contents must be reasonable and wreg contains
	 * the new scc configuration.  Disable rcvr & xmitter, load in new
	 * modes then re-enable.
	 */

	sc->sc_hs->hs_board[zs->zs_unit + HSIS_INT_CLK_A] = 0;
#ifdef HSIS_EXTERNAL_RCVDONE
	/*
	 * if there's an external receive clock, use the external
	 * done interrupt.  Otherwise, enable the chip's receive intr.
	 */
	if (sc->sc_sm.sm_rxclock == RXC_IS_RXC)
		sc->sc_hs->hs_board[zs->zs_unit + HSIS_EINT_ENA_A] = 0xff;
	else {
		sc->sc_hs->hs_board[zs->zs_unit + HSIS_EINT_ENA_A] = 0;
		wreg[1] |= ZSWR1_RIE_SPECIAL_ONLY;
	}
#endif
	SCC_BIC(3, ZSWR3_RX_ENABLE);
	SCC_BIC(5, ZSWR5_TX_ENABLE);

	SCC_WRITE(4, wreg[4]);
	SCC_WRITE(10, wreg[10]);
	SCC_WRITE(6, wreg[6]);
	SCC_WRITE(7, wreg[7]);
	SCC_WRITE(3, wreg[3]);
	SCC_WRITE(5, wreg[5]);
	SCC_WRITE(1, wreg[1]);
	SCC_WRITE(9, wreg[9]);
	SCC_WRITE(11, wreg[11]);
	SCC_WRITE(12, wreg[12]);
	SCC_WRITE(13, wreg[13]);
	if (clk == 2) {
		SCC_WRITE(14, ZSWR14_DPLL_SRC_BAUD);
		SCC_WRITE(14, ZSWR14_DPLL_NRZI);
	} else
		SCC_WRITE(14, ZSWR14_DPLL_DISABLE);
	SCC_WRITE(14, wreg[14]);
	SCC_WRITE(15, wreg[15]);

	if (txin) {
		/*
		 * TRxC pin is input (i.e., we're using an external clock).
		 * Set latch that allows clock to get to the pin.
		 */
		register u_char *board = sc->sc_hs->hs_board + zs->zs_unit;

		board[HSIS_INT_CLK_A] = 1;
		/* XXX - ext. clocks are inverted */
		board[HSIS_RX_CLK_A] = 1;
		board[HSIS_TX_CLK_A] = 1;
	}
	SCC_WRITE0(ZSWR0_RESET_ERRORS);
	SCC_WRITE0(ZSWR0_RESET_STATUS);
	SCC_WRITE0(ZSWR0_RESET_TXCRC);

	/*
	 * Everything should be configured.  Start a dma read then
	 * enable recv dma and the receiver.
	 */
	(void) hsis_start_dma_read(sc);
	SCC_BIS(1, ZSWR1_REQ_ENABLE)
	SCC_BIS(3, ZSWR3_RX_ENABLE);
	/*
	 * Enable the transmitter and turn on DTR.
	 */
	SCC_BIS(5, ZSWR5_TX_ENABLE|ZSWR5_DTR);

	SCC_WRITE0(ZSWR0_RESET_ERRORS);
	SCC_WRITE0(ZSWR0_RESET_STATUS);
	SCC_WRITE0(ZSWR0_RESET_TXCRC);

	sc->sc_ostate = 0;
	sc->sc_if.if_flags |= IFF_UP|IFF_RUNNING;
	TRACE(T_INIT_DONE, sc, *zsdev)
	splx(s);
	return (0);
}

hsis_reset(unit)
	register int unit;
{
	register int s = splboard();
	register struct hsis_softc *sc = &hsis_softc[unit];
	register struct zscc *zs = sc->sc_zs;
	register u_char *zsdev = zs->zs_addr;
	register u_char *zddev = sc->sc_zd->zd_addr;
	register struct hsisbuf *bp;

	TRACE(T_RESET, sc, *zsdev)
	SCC_WRITE(3, 0);
	SCC_WRITE(5, 0);
	SCC_WRITE(15, 0);
#ifdef HSIS_EXTERNAL_RCVDONE
	sc->sc_hs->hs_board[zs->zs_unit + HSIS_EINT_ENA_A] = 0;
#endif
	(void) hsis_start_dma_read(sc);
	if (zs->zs_unit & 1) {
		DMA_BIC(ZSDMA_DER, ZSDMA_DER_RX_B_ENABLE);
	} else {
		DMA_BIC(ZSDMA_DER, ZSDMA_DER_RX_A_ENABLE);
	}
	(void) hsis_start_dma_read(sc);
	if (sc->sc_intail) {
		sc->sc_intail->next = sc->sc_hs->hs_free;
		sc->sc_hs->hs_free = sc->sc_inq;
		sc->sc_inq = NULL;
		sc->sc_intail = NULL;
	}
	bzero(sc->sc_fifo, sizeof(sc->sc_fifo));
	sc->sc_fifop = sc->sc_fifo;
	sc->sc_inoff = NULL;

	sc->sc_ostate = 0;
	if (bp = sc->sc_curout)
		bp->cnt = 0;
	if (bp = sc->sc_nextout)
		bp->cnt = 0;
	sc->sc_if.if_flags &=~ (IFF_UP|IFF_RUNNING);
	splx(s);
}

hsis_offline(unit)
	register int unit;
{
	register int s = splboard();
	register struct hsis_softc *sc = &hsis_softc[unit];
	register struct hsisbuf *bp;

	/* reset channel, then free any resources it holds */

	hsis_reset(unit);
	if (bp = sc->sc_inbuf) {
		sc->sc_inbuf = NULL;
		bp->next = sc->sc_hs->hs_free;
		sc->sc_hs->hs_free = bp;
	}
	if (bp = sc->sc_curout) {
		sc->sc_curout = NULL;
		bp->next = sc->sc_hs->hs_free;
		sc->sc_hs->hs_free = bp;
	}
	if (bp = sc->sc_nextout) {
		sc->sc_nextout = NULL;
		bp->next = sc->sc_hs->hs_free;
		sc->sc_hs->hs_free = bp;
	}
	splx(s);
}

/*
 * Remainder of this code is support for `raw sync' protocol domain.
 * It should probably be in a separate file since it (should not be)
 * hsis specific but I'm too lazy to set that up just now.
 */

#include <sys/protosw.h>
#include <sys/domain.h>
#include <sys/user.h>
#include <sys/uio.h>

int rawsync_usrsend(), rawsync_usrrecv(), rawsync_usrreq();

extern struct domain rawsyncdomain;

struct protosw rawsyncsw[] = {
{ SOCK_RAW,	&rawsyncdomain,	0,		PR_ATOMIC,
  0,		0,		0,		0,
  rawsync_usrreq,
  0,		0,		0,		0,
  rawsync_usrsend, rawsync_usrrecv},
};

struct domain rawsyncdomain =
    { AF_RAWSYNC, "rawsync", 0, 0, 0,
      rawsyncsw, &rawsyncsw[sizeof(rawsyncsw)/sizeof(rawsyncsw[0])] };

void
rawsyncinit()
{
	register struct domain *dp;

	rawsyncdomain.dom_next = domains;
	domains = &rawsyncdomain;
}

int
rawsync_usrsend(so, nam, uio, flags, rights)
	register struct socket *so;
        struct mbuf *nam;
        register struct uio *uio;
        int flags;
        struct mbuf *rights;
{
        register int len;
        register int error = 0;
        register int s;
	register struct hsis_softc *sc;
	register struct mbuf *m;
	static struct sockaddr dst = { AF_UNSPEC };

#ifdef lint
        nam = nam; rights = rights;
#endif lint
	if ((sc = (struct hsis_softc *)so->so_pcb) == NULL ||
	    so != sc->sc_raw)
		return (EINVAL);

	len = uio->uio_resid;
	if (len <= 0 || len > sc->sc_if.if_mtu)
		return (EMSGSIZE);

	if (so->so_state & SS_CANTSENDMORE) {
		psignal(u.u_procp, SIGPIPE);
		return (EPIPE);
	}
	while (1) {
		s = splboard();
		if (! IF_QFULL(&sc->sc_if.if_snd))
			break;

		sbwait(&so->so_snd);
		splx(s);
	}
	(void) splnet();
	MGET(m, M_DONTWAIT, MT_DATA);
	if (m == (struct mbuf *)0) {
		error = ENOBUFS;
		goto out;
	}
	if (len <= MLEN - MBUF_HDRSPACE)
		m->m_off += MBUF_HDRSPACE;
	else {
		/* too big for mbuf - use cluster */
		MCLGET(m);
		if (m->m_len != MCLBYTES) {
			/* no clusters - drop this packet */
			m_freem(m);
			error = ENOBUFS;
			goto out;
		}
		m->m_off += HDRSPACE;
	}
	m->m_len = len;
	error = uiomove(mtod(m, caddr_t), len, UIO_WRITE, uio);
	if (! error)
		error = hsis_output(&sc->sc_if, m, &dst);
out:
	splx(s);
	return (error);
}

int
rawsync_usrrecv(so, anam, uio, flags, arights)
	register struct socket *so;
        struct mbuf **anam;
        register struct uio *uio;
        int flags;
        struct mbuf **arights;
{
        register int len, mlen;
        register int error = 0;
        register int s;
	register struct hsis_softc *sc;
	register struct mbuf *m, *n;
	register struct sockbuf *sb;

	if (anam)
		*anam = NULL;
	if (arights)
		*arights = NULL;

	if ((sc = (struct hsis_softc *)so->so_pcb) == NULL ||
	    so != sc->sc_raw)
		return (EINVAL);

	len = uio->uio_resid;
	if (len <= 0) {
		error = EMSGSIZE;
		goto out;
	}
	if (so->so_state & SS_CANTRCVMORE)
		goto out;

	sb = &so->so_rcv;
	sblock(so, sb);
	s = splboard();
	if (sb->sb_cc == 0) {
		if (so->so_error) {
			error = so->so_error;
			so->so_error = 0;
			goto release;
		}
		if (so->so_state & SS_NBIO) {
			error = EWOULDBLOCK;
			goto release;
		}
		while (sb->sb_cc == 0) {
			sbunlock(so, sb);
			sbwait(sb);
			if (so->so_error) {
				error = so->so_error;
				so->so_error = 0;
				goto release;
			}
		}
	}
	m = sb->sb_mb;
	sb->sb_mb = m->m_next;
	mlen = m->m_len;
	sb->sb_cc -= mlen;
	splx(s);
	sbunlock(so, sb);
	m->m_next = NULL;
	if (mlen > len) {
		error = EMSGSIZE;
		mlen = len;
	}
	error = uiomove(mtod(m, caddr_t), mlen, UIO_READ, uio);
	MFREE(m, n);
out:
	return (error);
release:
	splx(s);
	sbunlock(so, sb);
	return (error);
}

/*ARGSUSED*/
rawsync_usrreq(so, req, m, nam, rights)
	struct socket *so;
	int req;
	struct mbuf *m, *nam, *rights;
{
	register struct hsis_softc *sc;
	register int error = 0;

	if (rights && rights->m_len) {
		error = EOPNOTSUPP;
		goto release;
	}
	sc = (struct hsis_softc *)so->so_pcb;
	if (sc != NULL && so != sc->sc_raw) {
		error = EINVAL;
		goto release;
	}
	switch (req) {

	case PRU_ATTACH:
		if ((so->so_state & SS_PRIV) == 0)
			error = EACCES;
		else if (sc)
			error = EINVAL;
		else if (sbreserve(&so->so_snd, 4096) == 0)
			error = ENOBUFS;
		else if (sbreserve(&so->so_rcv, 4096) == 0) {
			sbrelease(&so->so_snd);
			error = ENOBUFS;
		}
		break;

	/*
	 * Destroy state just before socket deallocation.
	 * Flush data or not depending on the options.
	 */
	case PRU_DETACH:
		if (sc == 0)
			error = ENOTCONN;
		else {
			register int s = splboard(); 

			sc->sc_raw = NULL;
			so->so_pcb = NULL;
			sofree(so);
			splx(s);
		}
		break;

	case PRU_BIND:
		if (sc)
			error = EISCONN;
		else {
			register struct sockaddr *addr =
						mtod(nam, struct sockaddr *);
			if (addr->sa_family != AF_RAWSYNC) {
				error = EINVAL;
				break;
			}
			sc = (struct hsis_softc *)ifunit(addr->sa_data, MLEN);
			if (sc == NULL)
				error = EADDRNOTAVAIL;
			else if (sc->sc_raw)
				error = EADDRINUSE;
			else {
				register int s = splboard();

				sc->sc_raw = so;
				so->so_pcb = (caddr_t)sc;
				splx(s);
			}
		}
		break;

	/*
	 * Mark the connection as being incapable of further input.
	 */
	case PRU_SHUTDOWN:
		socantsendmore(so);
		break;

	case PRU_ABORT:
		if (sc != NULL && (so->so_state & SS_NOFDREF)) {
			register int s = splboard(); 

			sc->sc_raw = NULL;
			so->so_pcb = NULL;
			splx(s);
		}
		sofree(so);
		soisdisconnected(so);
		break;

	case PRU_SENSE:
		/*
		 * stat: don't bother with a blocksize.
		 */
		return (0);

	/*
	 * Not supported.
	 */
	case PRU_CONTROL:

	case PRU_CONNECT:
	case PRU_CONNECT2:
	case PRU_DISCONNECT:

	case PRU_RCVOOB:
	case PRU_RCVD:

	case PRU_LISTEN:
	case PRU_ACCEPT:
	case PRU_SENDOOB:

	case PRU_SOCKADDR:
	case PRU_PEERADDR:
		error = EOPNOTSUPP;
		break;

	default:
		panic("rawsync_usrreq");
	}
release:
	if (m != NULL)
		m_freem(m);
	return (error);
}

#endif
