#ifndef _ST2_API_H_
#define _ST2_API_H_ 1
#ifndef lint
static char rcsid_st2_api_h[] = "\
@(#)  $Header: st2_api.h,v 1.10 91/11/26 23:10:03 clynn Rel $  \n";
/*-------------------------------------------------------------------*
 |       Copyright (c) 1991 by BBN Systems and Technologies,         |
 |            A Division of Bolt Beranek and Newman Inc.             |
 |                                                                   |
 |  Permission to use, copy, modify, distribute, and sell this       |
 |  software and its documentation for any purpose is hereby         |
 |  granted without fee, provided that the above copyright notice    |
 |  and this permission appear in all copies and in supporting       |
 |  documentation, and that the name of Bolt Beranek and Newman      |
 |  Inc.  not be used in advertising or publicity pertaining to      |
 |  distribution of the software without specific, written prior     |
 |  permission. BBN makes no representations about the suitability   |
 |  of this software for any purposes.  It is provided "AS IS"       |
 |  without express or implied warranties.                           |
 *-------------------------------------------------------------------*/
#endif lint

/*
M* st2_api.h  ST-II Application Programming Interface Definitions.
M*
M*	This include file contains the Application Programming
M*	Interface Definitions for ST-II, an IP-layer protocol for
M*	experimentation with guaranteed quality of service, etc. (see
M*	RFC 1190 for a description or the ST-II Protocol).  It
M*	contains all definitions required by an ST-II Application, and
M*	then some.  These Application Interface Definitions are, as is
M*	ST-II, intended to provide flexibility and permit
M*	experimentation.  These definitions are for use with a
M*	BSD-style sockets interface.  To maximize transparency and
M*	extensibility, information crosses the OS/Application
M*	interface in network-byte order.  This file also contains some
M*	architecture and compiler specific datatype definitions.
M*
m* Status:
m*   Features:
m*	All the definitions that an ST-II Application should require.
m*	Configuration & Management definitions.
m*	Data Timestamps.
m*   Untested Features:
m*   Restrictions/Bugs:
m*	No support for Full-Duplex, Changes, full HID Negotiation,
m*	??? Notify, or Status.
m*   Things to do:
m*	Complete object descriptions.
m*	Add more InitXxx definitions.
m*	Add support for getsockopt/setsockopt.
m*
 */

/*
 * Experimenter Identification, to provide a challenge for the
 * monitoring software!  If you want to experiment with this software,
 * please get yourself a number (they are cheap, send a message to
 * stii-bugs@BBN.Com). If you have comments or suggestions, send them
 * there, too.
 *
 * Initial DARTNet experimenters ...
 */

#define ExperimenterList \
 aExp(0x00000000,ExpOfficial,stii-bugs@BBN.Com) \
 aExp(0x80000000,ExpBBN,CLynn@BBN.Com) \
 aExp(0x40000000,ExpISI,Casner@ISI.Edu) \
 aExp(0xC0000000,ExpLBL,) \
 aExp(0x20000000,ExpMIT,) \
 aExp(0xA0000000,ExpSRI,) \
 aExp(0x60000000,ExpUDel,) \
 aExp(0xE0000000,ExpUSC,) \
 aExp(0x10000000,ExpXerox,) \
 aExp(0x90000000,ExpINRIA,andry@purcell.inria.fr) \


#define Experimenter ExpOfficial	/* Your Experimenter Id goes here */

/* Module Revision History
 *
 *	bugid	4,0
 * $Log:	st2_api.h,v $
 * Revision 1.10  91/11/26  23:10:03  clynn
 * Made print format of config parameters and stats identical.
 * 
 * Revision 1.9  91/11/04  09:20:56  clynn
 * Updated for Public Domain Release.  Major changes: addition of Source
 * Routing, IP Encapsulation, HELLO protocol between neighbors, tracking
 * of neighbors, detection of component or agent failures & notification
 * to applications.  More consistant naming and format, including making
 * all external names begin with "st2_".  Moved some routines and data
 * structures to reduce external references.
 * 
 * Added more documentation and added missing pcode parameter to
 * InitaFlowSpec.
 * 
 * Revision 1.8  91/05/28  16:17:11  clynn
 * New features: Add new targets from Application layer, UserData support,
 * basic bandwidth reservation for point-to-point links, more complete
 * state tables, added pcode parameter to InitaFlowSpec3, extended packet
 * buffer abstraction, added network interface abstraction; ststat utility.
 * Bug fixes: ADDR IN USE problem, data send problem, causes of some
 * crashes, cleanup of protocol control blocks.
 * Work around: DARTNET receive memory leak.
 * Eliminated several small modules to reduce globals; adeded Makefile.
 * 
 * Revision 1.7  91/04/16  16:13:34  clynn
 * Comma missing in definition of InitaTarget.
 * 
 * Revision 1.1  91/03/15  18:33:05  clynn
 * Initial revision
 */


/* Definitions that should be in <sys/socket.h> ... */

#ifndef	AF_COIP				/* Defer to local site */
#define AF_COIP		23		/* COIP protocol family */
#endif	AF_COIP				/* see also ConfigParm (coip_family) */

#ifndef	PF_COIP
#define PF_COIP		AF_COIP		/* Same value, different semanitcs! */
#endif	PF_COIP


/* PF_COIP Definitions for ST (basically, same as for IP) */

#define STPROTO_ST	0		/* Dummy for ST */
/*#define STPROTO_TCP	IPPROTO_TCP	/* tcp */
/*#define STPROTO_UDP	IPPROTO_UDP	/* user datagram protocol */
#define STPROTO_RAW	255		/* raw ST packet */

#define IPPROTO_ST	5		/* Assigned Numbers */

/* From <netinet/in.h> ... */

#ifndef	INADDR_ANY
#define INADDR_ANY		(u_long)0x00000000
#endif	INADDR_ANY

#ifndef	s_addr				/* Defer to local site */
/*
 * Internet address
 */
struct in_addr {
    union {
	unsigned long	 S_addr;	/* An Internet Address */
    }			 S_un;
#define s_addr	S_un.S_addr
};

/*
 * Socket address, internet style.
 */
struct sockaddr_in {
    short		 sin_family;	/* I.E., AF_INET */
    unsigned short	 sin_port;	/* Transport-layer multiplexing info */
    struct in_addr	 sin_addr;	/* Identity of a system */
    unsigned char	 sin_zero[8];	/* Unused by AF_INET */
};
#endif	s_addr


/* Portability Definitions */
/* Machine dependent definitions */

#if defined(mc68020) || defined(mc68030) || defined(sparc)
typedef unsigned char	 octet1;
typedef unsigned short	 octet2;
typedef unsigned long	 octet4;
#define octet8 double			/* yea, but what's better?? */
#define NetOrder 1
#define ntoh1(x) x
#define hton1(x) x
#define ntoh2(x) x
#define hton2(x) x
#define ntoh4(x) x
#define hton4(x) x
#define ntoh8(x) x
#define hton8(x) x
#define ntohIP(x) ((INETADDR)(x))
#define htonIP(x) ((octet4)(x))
#endif


#if defined(mc68020) || defined(mc68030)
#define ALIGNMENT (1)
#endif defined(mc68020) || defined(mc68030)


#if defined(sparc)
#define ALIGNMENT (3)
#endif defined(sparc)

					/* Not to exceed ... */
#define MAX_ST_CTL	(1024)		/* form <sys/mbuf.h> MCLBYTES */
#define MAX_ST_NAM	(112)		/* from <sys/mbuf.h> MLEN */


/* Miscellaneous macros */

/* Split a number up into k bytes */
#define Bytes1(n) (n)
#define Bytes2(n) ((n)>>8),(0xff & (n))
#define Bytes3(n) ((n)>>16),(0xff & ((n)>>8)),(0xff & (n))
#define Bytes4(n) ((n)>>24),(0xff & ((n)>>16)),(0xff & ((n)>>8)),(0xff & (n))
/* others can be defined as required */


/* Build a symbol from three components */
#define Cat2(y,z)Ident(y)z


/* Build a comma-separated list */
#define Cma(x) x,


/* Find the dimension of an array */
#define DimensionOf(array) (sizeof (array) / sizeof (array[0]))


/* Read and write IP Addresses */
#define GetIPAdr(p,subscript) (unsigned long) (p)->IPAdr subscript
#define SetIPAdr(p,subscript,v) (p)->IPAdr subscript = (octet4) v


/* Build a symbol from two components */
#define Ident(x)x


/* Bulid an n-bit mask */
#define NBitMask(n) ((1 << (n)) - 1)


/* Round a number up modulo the size of a data type or structure. */
#define ROUND(n,t)	(((n) + sizeof (t) - 1) & ~ (sizeof (t) - 1))


/*
P* ST-II API Parameter Definitions.
P*
P*	To provide maximum flexibility and extensibility, we make the
P*	underlying ST-II Protocol definitions available to the Application.
P*	This should make it easier to add new Parameters and Messages for
P*	experimental purposes (plus eliminating reformatting by some layer);
P*	the dwonside is that the Application isn't isolated from changes in
P*	the underlying Protocols.
P*
 */


/* Maximum value of a Parameter's plen field, ((8*sizeof (plen))-1) & ~ 3 */
#define MaxPlen (252)


/*
p* Macros to build specific structures to be passed to ST by Application.
p*
p*	For each Parameter we define a structure prototype and
p*	structure initializer.  The prototypes may be customized by
p*	each Application when instantiated (at compile time), thus
p*	simplifying (really!) generation of parameters whose
p*	structure does not vary within an Application.  See the
p*	example program (example.c) for one way to use them.  NB:
p*	Since the order of parameters is not constrained by ST-II
p*	these prototypes cannot be used to "parse" parameters passed
p*	by ST to the Application!
p*
 */

/* ------------------------------------------------------------------------ */

/*
P* InstaApplEntity (name,sapbytes,srcruts)
p*
p* Structure to specify Application Entity Identifier pseudo-parameters
p* consisting of the tuple <IPAdr NextPcol SAPBytes SAP>.
p* PCodes: STLclAppEnt, STRmtAppEnt.
p* "name" is the optional name of the variable created.
p* "sapbytes" is the length of the SAP used by the Application (required).
p* "srcruts" is an optional list of Source Route Parameters (InstaSrcRut).
p*
 */
#define InstaApplEntity(name,sapbytes,srcruts)				\
  struct Ident (aApplEntity)name {					\
    octet1 pcode, plen, NextPcol, SAPBytes;				\
    octet4 IPAdr;							\
    octet1 SAP[(((sapbytes)+3)/4)*4];					\
    srcruts } name

/*
P* InitaApplEntity (type,self,ipadr,nextpcol,sapbytes,sap,srcruts)
p*
p* Initializer for an InstaApplEntity.
p* "type" is either STLclAppEnt or STRmtAppEnt.
p* "self" is the variable representing the structure, e.g., "name".
p* "ipadr" is the desired IP Address (or INADDR_ANY).
p* "nextpcol" identifies the protocol layer above ST.
p* "sapbytes" is the length of the SAP used by the Application (required).
p* "sap" is a comma-separated list of the bytes of the SAP (see Bytes2).
p* "srcruts" is a list of Source Route Initializers (InitaSrcRut).
p*
 */
#define InitaApplEntity(type,self,ipadr,nextpcol,sapbytes,sap,srcruts)	\
 { hton1 (type),hton1 (sizeof (self)),hton1 (nextpcol),			\
   hton1 (sapbytes), htonIP (ipadr),{sap},srcruts }

InstaApplEntity (,4,);		/* Generic definition for 1-4 byte SAPs */

/* ------------------------------------------------------------------------ */

/*
P* InstaDetectorIPAdr (name)
p*
p* Structure to specify a Detector IP Address Pseudo Parameter.
p* Passed to application in the control message structure.
p* PCodes: STDetectorIPAdr.
p* "name" is the optional name of the variable created.
p*
 */

#define InstaDetectorIPAdr(name) struct Ident (aDetectorIPAdr)name {	\
    octet1 pcode, plen;							\
    octet2 pad2;							\
    octet4 IPAdr; } name

InstaDetectorIPAdr ();			/* Generic Definition */

/* ------------------------------------------------------------------------ */

/*
P* InstaErrdPDU (name,pdubytes)
p*
p* Structure to specify a Errored PDU Parameter.
p* PCode: STpErrdPDU.
p* "name" is the optional name of the variable created.
p* "pdubytes" is the number of bytes to reserve for the Errored PDU.
p*
 */
#define InstaErrdPDU(name,pdubytes) struct Ident (aErrdPDU)name {	\
    octet1 pcode, plen,	 pdupbytes, erroroffset;			\
    octet1		 pduinerror[(((pdubytes)+3)/4)*4]; } name

InstaErrdPDU (,4);		/* Generic definition, not very useful */


/*
P* InstaFlowSpec (name)
p*
p* Structure to specify Resources required by an Application.
p* Multiple versions are expected during the course or the research.
p* All begin with a common 4-byte header which idenitfies the actual
p* version.  Ideally, only the Resource Management Module (see
p* st2_resource.c) actually needs to know the details of the specific
p* versions; we aren't quite there yet.
p* PCodes: STpFlowSpec, STpRFlowSpec.
p* "name" is the optional name of the variable created.
p*
 */
#define InstaFlowSpec(name) struct Ident (aFlowSpec)name {		\
    octet1 pcode, plen, FlowVer, fill1; } name

InstaFlowSpec ();			/* Generic header */

/* ------------------------------------------------------------------------ */

/*
P* InstaFlowSpec3 (name)
p*
p* Structure to specify a Version 3 FlowSpec Parameter.
p* RFC 1190 defines the fields in FlowSpec Version 3.
p* "name" is the optional name of the variable created.
p*
 */
#define STFSVer3	(3)		/* Version of RFC 1190 */
#define STFSImpAll	(0x8000)
#define STFSUnknown	(0x4000)
#define STFSRevChrg	(0x2000)

#define InstaFlowSpec3(name) struct Ident (aFlowSpec3)name {		\
    octet1 pcode, plen, FlowVer, fill1, DutyFactor, ErrorRate,		\
	   Precedence, Reliability;					\
    octet2 Tradeoffs, RecoveryTimeout, LimitOnCosts, LimitOnDelay,	\
	   LimitOnPDUBytes, LimitOnPDURate;				\
    octet4 MinBytesXRate, AccdMeanDelay, AddDelayVariance;		\
    octet2 DesPDUBytes, DesPDURate; } name

/*
P* InitaFlowSpec3 (pcode,minbytes,minpps,minbw,desbytes,despps)
p*
p* Initializer for a Version 3 Flowspec for use by Applications
p* only interested in specifying packet size and rate information.
p* "minbytes" is the minimum acceptable number of bytes per packet.
p* "minpps" is the minimum acceptable number of Data packets per second.
p* "minbw" is the minimum acceptable bandwidth (bytes * packets/second).
p* "desbytes" is the (largest useful) desireable number of bytes per packet.
p* "despps" is the (largest useful) desireable number of packets/second.
p*
 */
#define APIRecoveryTimeout (0)		/* msec (default) */
#define APILimitOnDelay    (2000)	/* msec */

#define InitaFlowSpec3(pcode,minbytes,minpps,minbw,desbytes,despps)	\
 { hton1 (pcode),hton1 (sizeof (struct aFlowSpec3)),hton1 (STFSVer3),	\
   0,0,0,0,0,0,hton2 (APIRecoveryTimeout),0,hton2 (APILimitOnDelay),	\
   hton2 ((minbytes)),hton2 (((minpps)*10)),hton4 (((minbw)*10)),0,0,	\
   hton2 ((desbytes)),hton2 (((despps)*10)) }

InstaFlowSpec3 ();		/* Generic Version 3 FlowSpec */

/* ------------------------------------------------------------------------ */

/*
P* InstaFreeHIDs (name,nmasks)
p*
p* Structure to specify a Free HIDs Parameter.
p* PCodes: STpFreeHIDs.
p* "name" is the optional name of the variable created.
p* "nmasks" is the number of 32-bit HID mask words allocated (required).
p*
 */
#define InstaFreeHIDs(name,nmasks) struct Ident (aFreeHIDs)name {	\
    octet1 pcode, plen;							\
    octet2 basehid;							\
    octet4 freehidbitmask[ nmasks ]; } name

InstaFreeHIDs (,2);		/* Generic FreeHIDs w/ 64-bits */

/* ------------------------------------------------------------------------ */

/*
P* InstaGroup (name,nsubgroups)
p*
p* Structure to specify a Group or Reverse Group Parameter.
p* PCodes: STpGroup, STpRGroup.
p* "name" is the optional name of the variable created.
p* "nsubgroups" specifies the number of sub-group slots allocated (required).
p*
 */
#define InstaGroup(name,nsubgroups) struct Ident (aGroup)name {		\
    octet1 pcode, plen;							\
    octet2 UniqId;							\
    octet4 IPAdr, Timstmp;						\
    struct msubgroup							\
    { octet2 subgroupid, relation;					\
    } subgroups[nsubgroups]; } name

InstaGroup (,1);			/* Generic Group/RGroup Parameter */

/* ------------------------------------------------------------------------ */

/*
P* InstaHID (name)
p*
p* Structure to specify a HID or Reverse HID Parameter.
p* An SCM Notify or StatusResponse PDU may contain multiple HID Parameters.
p* PCodes: STpHID, STpRHID.
p* "name" is the optional name of the variable created.
p*
 */
#define InstaHID(name) struct Ident (aHID)name {			\
    octet1 pcode, plen;							\
    octet2 hid; } name

InstaHID ();			/* Generic HID Parameter */

/* ------------------------------------------------------------------------ */

/*
P* InstaMcastAdr (name,mlen)
p*
p* Structure to specify a Multicast Address Parameter.
p* PCodes: STpMcastAdr.
p* "name" is the optional name of the variable created.
p* "mlen" is the number of bytes in the local net layer multicast address
p*	  (required).
p*
 */
#define InstaMcastAdr(name,mlen) struct Ident (aMcastAdr)name {		\
    octet1 pcode, plen, localnetbytes, fill1;				\
    octet4 ipmulticastaddress;						\
    octet1 localnetmulticastaddress[ (((mlen)+3)/4)*4 ]; } name

InstaMcastAdr (,6);		/* McastAdr Parameter for Ethernets */

/* ------------------------------------------------------------------------ */

/*
P* InstaName (name)
p*
p* Structure to specify a Name or Reverse Name Parameter.
p* PCodes: STpName, STpRName.
p* "name" is the optional name of the variable created.
p*
 */
#define InstaName(name) struct Ident (aName)name {			\
    octet1 pcode, plen;							\
    octet2 UniqId;							\
    octet4 IPAdr, Timstmp; } name

InstaName ();			/* Generic Name/RName Parameter */

#define sizeofaName (sizeof (struct aName))

/* ------------------------------------------------------------------------ */

/*
P* InstaNxtHopIPAdr (name)
p*
p* Structure to specify a Next-Hop IP Address Parameter.
p* Used with the Notify SCMP message to identify a potential routing problem.
p* PCodes: STpNxtHopIPAdr.
p* "name" is the optional name of the variable created.
p*
 */
#define InstaNxtHopIPAdr(name) struct Ident (aNxtHopIPAdr)name {	\
    octet1 pcode, plen;							\
    octet4 IPAdr; } name

InstaNxtHopIPAdr ();		/* Generic NxtHopIPAdr Parameter */

/* ------------------------------------------------------------------------ */

/*
P* InstaOrigin (name,sapbytes)
p*
p* Structure to specify a Origin Parameter.
p* PCodes: STpOrigin.
p* "name" is the optional name of the variable created.
p* "sapbytes" is the length of the SAP used by the Application (required).
p*
 */
#define InstaOrigin(name,sapbytes) struct Ident (aOrigin)name {		\
    octet1 pcode, plen, NextPcol, SAPBytes;				\
    octet4 IPAdr;							\
    octet1 SAP[ (((sapbytes)+3)/4)*4 ]; } name

InstaOrigin (,4);		/* Generic Origin for 1-3 byte SAPs */

/* ------------------------------------------------------------------------ */

/*
P* InstaOrgTsmp (name)
p*
p* Structure to specify a Origin Timestamp Parameter.
p* PCodes: STpOrgTsmp.
p* "name" is the optional name of the variable created.
p*
 */
#define InstaOrgTsmp(name) struct Ident (aOrgTsmp) {			\
    octet1 pcode, plen;							\
    octet2 fill2;							\
    octet4 timesec, fractsec; } name

InstaOrgTsmp ();			/* Generic Origin Timestamp Param */

/* ------------------------------------------------------------------------ */

/*
P* InstaParameter (name)
p*
p* Structure to specify a Generic Parameter.
p* PCodes: <all>.
p* "name" is the optional name of the variable created.
p*
 */
#define InstaParameter(name) struct Ident (aParameter)name {		\
    octet1 pcode, plen;							\
    octet2 pdata2;							\
    octet4 pdata4; } name

InstaParameter ();		/* Generic Parameter */

/* ------------------------------------------------------------------------ */

/*
P* InstaReasonCode (name)
p*
p* Structure to specify a Reason Code Pseudo Parameter.
p* Reason (failure) codes are communicated between the Application and
p* ST using this pseudo parameter.
p* PCodes: STReasonCode.
p* "name" is the optional name of the variable created.
p*
 */

#define InstaReasonCode(name) struct Ident (aReasonCode)name {		\
    octet1 pcode, plen;							\
    octet2 reason; } name

/*
P* InitaReasonCode (reason)
p*
p* Initializer for a Reason Code Pseudo Parameter.
p* "reason" is the appropriate enum ST2Errors symbol.
p*
 */
#define InitaReasonCode(reason)						\
 { hton1 (STReasonCode),hton1 (sizeof (struct aReasonCode)),		\
   hton2 (((octet2) reason)) }

InstaReasonCode ();		/* Generic Reason Code Parameter */

/* ------------------------------------------------------------------------ */

/*
P* InstaRecordRoute (name,naddrs)
p*
p* Structure to specify a Record Route Parameter.
p* PCodes: STpRecordRoute.
p* "name" is the optional name of the variable created.
p* "naddrs" specifies the number of IP Address slots to be allocated
p*	    (required).
p*
 */
#define InstaRecordRoute(name,naddrs) struct Ident (aRecordRoute)name {	\
    octet1 pcode, plen, fill1, freeoffset;				\
    octet4 IPAdr[naddrs]; } name

InstaRecordRoute (,1);		/* Generic Record Route Parameter */

/* ------------------------------------------------------------------------ */

/*
P* InstaSrcRut (name,naddrs)
p*
p* Structure to specify a Source Route Pseudo Parameter.
p* These pseudo parameters is used within a Target specification to
p* specify source routing information.  They may appear multiply and
p* in any combination per Target.
p* PCodes: STpIPLSrcRut, STpIPSSrcRut, STpSTLSrcRut, STpSTSSrcRut.
p* "name" is the optional name of the variable created.
p* "naddrs" specifies the number of IP Address slots to be allocated.
p*
 */
#define InstaSrcRut(name,naddrs) struct Ident (aSrcRut)name {		\
    octet1 pcode, plen;							\
    octet2 fill2;							\
    octet4 IPAdr[naddrs]; } name

/*
P* InitaSrcRut (type,self,ipaddrs)
p*
p* Initializer for a Source Route Parameter.
p* "type" is one of STpIPLSrcRut, STpIPSSrcRut, STpSTLSrcRut, or STpSTSSrcRut.
p* "self" is the variable representing the structure, e.g., "name".
p* "ipaddrs" is a list of "Cma(" IP Address ")" es forming the route.
p*
 */
#define InitaSrcRut(type,self,ipaddrs)					\
 { hton1 (type),hton1 (sizeof (self)),0,{ipaddrs} }

InstaSrcRut (,1);		/* Generic Source Route Parameter */
				/* NB: sizeof () is minimum valid size */

/* ------------------------------------------------------------------------ */

/*
P* InstaTarget (name,sapbytes,srcruts)
p*
p* Structure to specify a Target, used within a Target List Parameter.
p* "name" is the optional name of the variable created.
p* "sapbytes" is the length of the SAP used by the Application (required).
p* "srcruts" is an optional list of Source Route Parameters (InstaSrcRut).
p*
 */
#define InstaTarget(name,sapbytes,srcruts) struct Ident (aTarget)name {	\
    octet4 IPAdr;							\
    octet1 TargetBytes, SAPBytes;					\
    octet1 SAP[((((sapbytes)+1)/4)*4)+2];				\
    srcruts } name

/*
P* InitaTarget (self,ipadr,sapbytes,sap,srcruts)
p*
p* Initializer for a Target Specification.
p* "self" is the variable representing the structure, e.g., "name".
p* "ipadr" is the desired IP Address (or INADDR_ANY).
p* "sapbytes"is the length of the SAP used by the Application (required).
p* "sap" is a comma-separated list of the bytes of the SAP (see Bytes2).
p* "srcruts" is a list of Source Route Initializers (InitaSrcRut).
p*
 */
#define InitaTarget(self,ipadr,sapbytes,sap,srcruts)			\
 { htonIP ((ipadr)),hton1 (sizeof (self)),hton1 ((sapbytes)),{sap},srcruts },

InstaTarget (,2,);		/* Generic Target for 1-2 byte SAPs */

#define minsizeofaTarget (8)

/* ------------------------------------------------------------------------ */

/*
P* InstaTargetList (name,targets)
p*
p* Structure to specify a Target List Parameter.
p* PCodes: STpTargetList.
p* "name" is the optional name of the variable created.
p* "targets" is a list of Target Specifications (InstaTarget).
p*
 */
#define InstaTargetList(name,targets) struct Ident (aTargetList)name {	\
    octet1 pcode, plen;							\
    octet2 TargetCount;							\
    targets } name

/*
P* InitaTargetList (self,ntargets,targets)
p*
p* Initializer for a Target List Parameter.
p* "self" is the variable representing the structure, e.g., "name".
p* "ntargets" specifies the number of Target Specifications in the List.
p* "targets" is a list of Target Specification Initializers (InitaTarget).
p*
 */
#define InitaTargetList(self,ntargets,targets)				\
 { hton1 (STpTargetList),hton1 (sizeof (self)),hton2 (ntargets),targets }

InstaTargetList (,struct aTarget targets[1];); /* Generic Single Target */

InstaTargetList (header,)aTargetList;	/* Generic header w/o any Targets */
#define sizeofaTargetListHeader (sizeof (struct aTargetListheader))

/* ------------------------------------------------------------------------ */

/*
P* InstaUserData (name,bytes)
p*
p* Structure to specify a User Data Parameter.
p* It may be used by an Application to communicate setup(or teardown)
p* information between the Origin and Targets of a Stream.
p* PCodes: STpUserData.
p* "name" is the optional name of the variable created.
p* "bytes" is the number of bytes of space to be allocated.
p*
 */
#define InstaUserData(name,bytes) struct Ident (aUserData)name {	\
    octet1 pcode, plen;							\
    octet2 userlen;							\
    octet1 userdata[((bytes+3)/4)*4]; } name

InstaUserData (,4);		/* Generic User Data Parameter, 1-3 bytes */

/* ------------------------------------------------------------------------ */

/*
P* Instsockaddr_st2 (name,st_parms)
p*
p* Structure to specify an ST-II sockaddr.
p* ST Requests (enum STRequests) are specified implicitly with certain
p* BSD-style socket calls (bind, connect, listen, accept) or explicitly
p* in the ST-II sockaddr (and cmsghdr) structures.
p* "name" is the optional name of the variable created.
p* "st_parms" is a list of ST Parameters (InstXxx).
p*
 */
#define Instsockaddr_st2(name,st_parms) struct Ident (sockaddr_st2)name { \
    unsigned short	 sa_family;	/* AF_COIP */			\
    unsigned char	 st_options;	/* enum STOptions */		\
    unsigned char	 st_request;	/* enum STRequests */		\
    unsigned long	 st_handle;	/* ffs */			\
    unsigned long	 st_sec, st_fract; /* NTP-format Data Timestamp */ \
    st_parms } name

/*
P* Initsockaddr_st2 (req,opt)
p*
p* Initializer for a Target List Parameter.
p* "req" is an element of enum STRequests.
p* "opt" is an "|"-separated list of ST Options (enum STOptions).
p*
 */
#define Initsockaddr_st2(req,opt) AF_COIP,opt,req,0,0,0

Instsockaddr_st2 (,struct aParameter st_parms[1];); /* Generic ST sockaddr */

Instsockaddr_st2 (header,)sockaddr_st2;	/* Generic header */
#define sizeofsockaddr_st2header (sizeof (struct sockaddr_st2header))

/* ------------------------------------------------------------------------ */

/*
O* ST Options.
o*
o*	Options may appear in the st_options field of a sockaddr_st2 structure.
o*
 *	For each field/bit combination, we define the bit mask, name, and
 *	description of the Option.
 *
 *	aSTOption (value,name,description)
 */

#define STOptionList \
 aSTOption (0x80,STOpt1socket,Bind (no support)) \
 aSTOption (0x40,STOptPBit,CONNECT/ACCEPT PTP option) \
 aSTOption (0x20,STOptSBit,CONNECT/ACCEPT NoRecovery) \
 aSTOption (0x03,STTSR,CONNECT/ACCEPT Data Timestamp) \
 aSTOption (0x00,STTSRNotImp,CONNECT/ACCEPT) \
 aSTOption (0x01,STTSRNo,CONNECT/ACCEPT) \
 aSTOption (0x02,STTSRYes,CONNECT/ACCEPT) \
 aSTOption (0x03,STTSROpt,CONNECT/ACCEPT) \
 aSTOption (0xE0,STDataPriority,Data priority send/recv) \
 aSTOption (0x10,STDataTBit,Data Timestamp send/recv) \


enum STOptions {
#define aSTOption(bits,name,desc) name = bits,
    STOptionList
    syntaxfixer
#undef aSTOption
};

/* ------------------------------------------------------------------------ */

/*
O* ST Requests.
o*
o*	Requests may appear in the st_request field of a sockaddr_st2 structure
o*	or the cmsg_type field of a cmsghdr structure.
o*
 *	For each Request or Message OpCode, we define the value, name, and
 *	description of the Request.  (Not all SCMP Messages are directly
 *	available to the Application Layer.)
 *
 *	aReq (value,name,comment)
 *
 *	ApplicationSpecificRequests is proviced to allow applications to
 *	insert additional (non-ST-II) elements in the enumeration, if desired.
 */

#define STRequestList \
 aReq(0,STReqUnspec,App->ST request as per function or ST->App No notification or data) \
 aReq(1,STReqAccept,App->ST ACCEPT connection or ST->App CONNECT received) \
 xReq(2,Ack,) \
 aReq(3,STReqChange,CHANGE  sent/received) \
 aReq(4,STReqChangeRequest,App->ST request origin to make change or ST->App requested change) \
 aReq(5,STReqConnect,App->ST send CONNECT) \
 aReq(6,STReqDisconnect,App->ST send DISCONNECT) \
 xReq(7,ErrorInRequest,) \
 xReq(8,ErrorInResponse,) \
 xReq(9,Hello,) \
 xReq(10,HIDApprove,) \
 xReq(11,HIDChange,) \
 xReq(12,HIDChangeRequest,) \
 xReq(13,HIDReject,) \
 aReq(14,STReqNotify,NOTIFY  sent/received) \
 aReq(15,STReqRefuse,App->ST send REFUSE) \
 aReq(16,STReqStatus,Request for status) \
 aReq(17,STReqStatusResponse,Status response) \
 xReq(18,x18,) \
 xReq(19,x19,) \
 xReq(20,x20,) \
 xReq(21,x21,) \
 xReq(22,x22,) \
 xReq(23,x23,) \
 xReq(24,x24,) \
 xReq(25,x25,) \
 xReq(26,x26,) \
 xReq(27,x27,) \
 xReq(28,x28,) \
 xReq(29,x29,) \
 xReq(30,x30,) \
 xReq(31,x31,) \
 aReq(32,STCtlPending,Control information is waiting for application) \
 aReq(33,STReqBind,App->ST bind () ) \
 aReq(34,STReqGroupName,Request a unique GroupName) \
 aReq(35,STReqGroupRel,Release a GroupName) \
 aReq(36,STReqParms,App->ST specify parametrers) \


#define MAX_ST_REQUEST (36+1)

#ifndef ApplicationSpecificRequests
#define ApplicationSpecificRequests
#endif

enum STRequests {
#define aReq(value,name,comment) name = value,
#define xReq(value,name,comment)
    STRequestList
    ApplicationSpecificRequests
#undef xReq
#undef aReq
    NUM_ST_REQUESTS
};


/* SCM Protocol definitions */

/* Define each SCMP Message (OpCode) object.  For each object, we specify the
 *   value
 *   name
 *   QB(option bit)
 *   QM(mandatory parameter)
 *   QO(optional parameter)
 *   QV(r/s VLId required)
 * as per the ST-II Protocol Specification (RFC 1190).
 * Due to C deficiencies, must be in numberic order, starting at 0, w/o any
 * gaps.
 */

/* ??? Add properties of object, */
/*     e.g., permit 0 rvlid (Connect, Hello, Status, ?Notify) */
/* ??? Option bits (OB ()), required recv VLIds */
/* ??? add Query/Response, ..., merge api STReq* */


#define ST2OpCodes \
 anOpCode(0,BadOpCode,,,,) \
 anOpCode(1,Accept,,QM(STDetectorIPAdr) QM(STpName) QM(STpFlowSpec) \
			QM(STpTargetList), \
		QO(STpRecordRoute) QO(STpRFlowSpec) QO(STpRName) \
			QO(STpUserData), \
		QV(R) QV(S)) \
	/* HELLOs don't have Names, so neither can their ACKs */ \
 anOpCode(2,Ack,,QM(STReasonCode),QO(STpName),QV(R) QV(S)) \
 anOpCode(3,Change,QB(G),QM(STDetectorIPAdr) QM(STpName) QM(STpFlowSpec), \
		QO(STpTargetList) QO(STpUserData), \
		QV(R) QV(S)) \
 anOpCode(4,ChangeRequest,QB(G),QM(STDetectorIPAdr) QM(STpName) \
			QM(STpFlowSpec), \
		QO(STpTargetList) QO(STpUserData), \
		QV(R) QV(S)) \
 anOpCode(5,Connect,QB(H) QB(P) QB(S) QB(TSP), \
		QM(STDetectorIPAdr) QM(STpName) QM(STpOrigin) QM(STpFlowSpec) \
		QM(STpTargetList),QO(eHID) QO(STpGroup) QO(STpMcastAdr) \
		QO(STpRecordRoute) QO(STpRFlowSpec) QO(STpRGroup) QO(STpRHID) \
		QO(STpUserData), \
		QV(S)) \
 anOpCode(6,Disconnect,QB(G),QM(STReasonCode) QM(STDetectorIPAdr) QM(STpName), \
		QO(STpTargetList) QO(STpUserData), \
		QV(R) QV(S)) \
 anOpCode(7,ErrorInRequest,,QM(STReasonCode) QM(STDetectorIPAdr), \
		QO(STpName) QO(STpErrdPDU) QO(STpTargetList), \
		QV(R)) \
 anOpCode(8,ErrorInResponse,,QM(STReasonCode) QM(STDetectorIPAdr), \
		QO(STpName) QO(STpErrdPDU) QO(STpTargetList), \
		QV(R) QV(S)) \
 anOpCode(9,Hello,QB(R),QM(eHelloTimer),QO(STpOrgTsmp),QV(S)) \
 anOpCode(10,HidApprove,,QM(eHID) QM(STpName),QO(STpFreeHIDs), \
		QV(R) QV(S)) \
 anOpCode(11,HidChange,QB(A) QB(D),QM(eHID) QM(STpName),, \
		QV(R) QV(S)) \
 anOpCode(12,HidChangeRequest,QB(A) QB(D),QM(eHID) QM(STpName),, \
		QV(R) QV(S)) \
 anOpCode(13,HidReject,,QM(eHID) QM(STpName),QO(STpFreeHIDs), \
		QV(R) QV(S)) \
 anOpCode(14,Notify,,QM(STReasonCode) QM(STDetectorIPAdr),QO(STpErrdPDU) \
		QO(STpFlowSpec) QO(STpHID) QO(STpName) QO(STpNxtHopIPAdr) \
		QO(STpRecordRoute) QO(STpTargetList),) \
 anOpCode(15,Refuse,,QM(STReasonCode) QM(STDetectorIPAdr) QM(STpName) \
			QM(STpTargetList), \
		QO(STpErrdPDU) QO(STpRecordRoute) QO(STpUserData), \
		QV(R) QV(S)) \
 anOpCode(16,Status,QB(H) QB(Q),QM(eHID),QO(STpRName),QV(S)) \
 anOpCode(17,StatusResponse,QB(H) QB(Q),QM(eHID) QM(STpRName), \
		QO(STpFlowSpec) QO(STpGroup) QO(STpHID) QO(STpTargetList), \
		QV(R)) \


/* ??? add connect-request */

#ifdef DOCUMENTATION

a) q - ErrorInRequest, r - ErrorInReply, n - no error
b) a - DoAcknowledge
c) has rx parameters
d) wants a valid vlinkp
e) f- notify_pending_Forwarder, t - notify_pending_Target,
   d - disconnect_Forwarder, all disconect_Target
f) st2_PCBUpdate
g) relay message, q - queue_accepts

		|   a	|   b	|   c	|   d	|   e	|   f	|   g	|
		+-------+-------+-------+-------+-------+-------+-------+
      BadOpCode	|   q?	|	|	|	|	|	|	|
	 Accept	|   q	|   a	|   y	|   na	|   t	|	|  y/q	|
	    Ack	|   r	|	|	|   y	|   f	|   y	|   n	|
	 Change	|   q	|	|   +2	|   y	|	|	|	|
  ChangeRequest	|   q	|	|   +1	|   y	|	|	|	|
	Connect	|   q	|	|   y	|	|	|	|	|
     Disconnect	|   q	|   a	|   y	|   y	|   d	|   y	|	|
 ErrorInRequest	|   n	|	|	|   y	|   f	|   y	|   n	|
ErrorInResponse	|   n	|	|	|   y	|   f	|   y	|   n	|
	  Hello	|   q	|	|	|	|	|	|	|
     HidApprove	|   r	|	|	|   y	|   f	|   y	|	|
      HidChange	|   q	|	|   y	|	|	|	|	|
HidChangeRequest|   q	|	|   -1	|	|	|	|	|
      HidReject	|   r	|	|	|   y	|   f	|   y	|	|
	 Notify	|   q	|	|   y	|	|	|	|	|
	 Refuse	|   q	|   a	|   y	|   y	|   ?	|   ?	|   y	|
	 Status	|   q	|	|	|	|	|	|	|
 StatusResponse	|   q	|	|	|	|	|	|	|

#endif DOCUMENTATION

enum OpCodes {
#define anOpCode(value,name,bits,mparm,oparm,vlids) Ident (Op)name = value,
    ST2OpCodes
#undef anOpCode

    NUM_ST_OPCODES
};

/* ------------------------------------------------------------------------ */

/* Define each PCode object.
 *	For each object, we specify the value, structure prefix, enum
 *	prefixm, name, min and max lengths, and <more later>, as per
 *	the ST-II spec.
 *
 *	aPCode (value,struct-prefix,pcode-prefix,(base)name,minlen,maxlen,x)
 *	sPCode - source routes, tPCode - targetlist, xPCode - internal
 */

#define ST2PCodes \
 xPCode(0,x,p,Bogus,0,0,Catch unspecified parameter code) \
 aPCode(1,a,STp,ErrdPDU,4,252,Copy of Errored SCM PDU) \
 aPCode(2,a,STp,FlowSpec,36,36,FlowSpec describing Resource Requirements) \
 aPCode(3,a,STp,FreeHIDs,8,36 \
		,Masks of HIDs which are available at the sender) \
 aPCode(4,a,STp,Group,12,28,Information specifying a Group of Streams) \
 aPCode(5,a,STp,HID,4,4 \
		,Identifies HID in use by a stream between two ST Agents) \
 aPCode(6,a,STp,McastAdr,8,20,Communicate multicast address information) \
 aPCode(7,a,STp,Name,12,12,Identify a Stream) \
 aPCode(8,a,STp,NxtHopIPAdr,8,8,Communicate a Next Hop IP Address) \
 aPCode(9,a,STp,Origin,8,16,Identify the Origin of a Stream) \
 aPCode(10,a,STp,OrgTsmp,12,12,Communicate Time at Origin when Data was sent) \
 aPCode(11,a,STp,RecordRoute,4,252 \
		,Parameter to collect stream path information) \
 aPCode(12,a,STpR,FlowSpec,36,36 \
		,FlowSpec describing reverse Resource Requirements) \
 aPCode(13,a,STpR,Group,12,28 \
		,Information specifying a reverse Group of Streams) \
 aPCode(14,a,STpR,HID,4,4,Specify HID to be used by Reverse Stream) \
 aPCode(15,a,STpR,Name,12,12,Identify a Reverse Stream) \
 sPCode(16,a,STpIPL,SrcRut,4,252,Loose IP Source Route) \
 sPCode(17,a,STpIPS,SrcRut,4,252,Strict IP Source Route) \
 sPCode(18,a,STpSTL,SrcRut,4,252,Loose ST Source Route) \
 sPCode(19,a,STpSTS,SrcRut,4,252,Strict ST Source Route)  \
 tPCode(20,a,STp,TargetList,12,252,List of Targets for a Stream) \
 aPCode(21,a,STp,UserData,4,252 \
		,Communicate Application Data at Setup/Teardown) \
 xPCode(22,no,x,22,0,0,x) \
 xPCode(23,no,x,23,0,0,x) \
 xPCode(24,no,x,24,0,0,x) \
 xPCode(25,aApplEntity,STLclAppEnt,,8,252 \
		,Local Application Entity <IPAdr NextPcol SAPBytes SAP>) \
 xPCode(26,aApplEntity,STRmtAppEnt,,8,252 \
		,Remote Application Entity <IPAdr NextPcol SAPBytes SAP>) \
 xPCode(27,no,e,HelloTimer,4,4,x) \
 xPCode(28,no,e,HID,2,2,x) \
 xPCode(29,a,ST,DetectorIPAdr,4,4,x) \
 xPCode(30,a,ST,ReasonCode,2,2,x) \


/* ??? add accounting info */
/* ??? add stats info */

enum PCodes {
#define aPCode(v,s,p,n,mn,mx,x) Ident (p)n = v,
#define sPCode(v,s,p,n,mn,mx,x) Ident (p)n = v,
#define tPCode(v,s,p,n,mn,mx,x) Ident (p)n = v,
#define xPCode(v,s,p,n,mn,mx,x) Ident (p)n = v,
    ST2PCodes
#undef xPCode
#undef tPCode
#undef sPCode
#undef aPCode

    NUM_ST_PARMS
};


/*
 * Level number for [gs]etsockopt () to apply to ST.
 */

#define SOL_STII	(1190)		/* ST-II (RFC 1190) option level */

/*
 * [gs]etsockopt () options
 *
 *	NB: ST socket option codes are an extension of the user request
 *	function (pr_usrreq in struct protosw) codes (PRU_xxx) defined
 *	in <sys/protosw.h>.  A gap is left after PRU_NREQ for growth.
 *	(*protosw[].pr_usrreq)(so, PRU_xxx, m, nam[sockaddr], opt);
 *
 *	getsockopt ( int s, level, option, caddr_t val, int *vallen )
 *	setsockopt ( int s, level, option, caddr_t val, int vallen )
 *	However, they use the pr_ctloutput function.
 *	(*pr_ctloutput)( PRCO_[GF]SETOPT, so, level, opt, struct mbuf **m0 )
 *	returning a UNIX errno.
 */

#define STOptPriority	(65)

/* ------------------------------------------------------------------------ */

/*
 * Control structure used with recvmsg () and sendmsg () and [gs]etsockopt ()
 */

/*
P* Instcmsghdr (name,parms)
p*
p* Structure to specify a Control Message Header.
p*
p*	Used with getsockopt and setsockopt as well as forming the contents
p*	of the msg_accrights field of the msghdr structure used with
p*	recvmsg and snedmsg system calls.
p*
p* "name" is the optional name of the variable created.
p* "parms" is a list of ST Parameter specifications (InstaXxx).
p*
 */
#define Instcmsghdr(name,parms) struct Ident (cmsghdr)name {		\
    unsigned int	 cmsg_len;					\
    int			 cmsg_level, cmsg_type;				\
    parms } name

/*
P* Initcmsghdr (self,request)
p*
p* Initializer for a Control Message Header.
p* "self" is the variable representing the structure, e.g., "name".
p* "request" is the appropriate enum STRequest symbol.
p*
 */
#define Initcmsghdr(self,request) sizeof (self),SOL_STII,request

Instcmsghdr (,unsigned char cmsg_data[4];); /* Generic cmsghdr */

Instcmsghdr (header,)cmsghdr;		/* Generic Header only */
#define sizeofCMsgHdr (sizeof (struct cmsghdrheader))

/* ------------------------------------------------------------------------ */

/*
P* InitMsg (p,name,namelen,iov,niov,ctl,ctllen)
p*
p* The recvmsg () and sendmsg () calls use the struct msg structure.
p* "p" is a pointer to the struct msg.
p* "name" is a pointer to a struct sockaddr_st2.
p* "namelen" is the length of the sockaddr_st2.
p* "iovp" is a pointer to a struct iov array.
p* "niov" is the number of entries in the struct iov array.
p* "ctlp" is a pointer to a struct cmsghdr containing a Parameter list.
p* "ctllen" is the length of the cmsghdr and Parameters.
p*
 */

#define InitMsg(p,name,namelen,iovp,niov,ctlp,ctllen)			\
    (p)->msg_name = (caddr_t) (name);					\
    (p)->msg_namelen = (namelen);					\
    (p)->msg_iov = (struct iovec *) (iovp);				\
    (p)->msg_iovlen = (niov);						\
    (p)->msg_accrights = (caddr_t) (ctlp);				\
    (p)->msg_accrightslen = (ctllen);


/*
 * Error Handling
 *
 *	Needs more work :-)
 *	What do you do if you DETECT an error
 *	What do you do if you RECEIVE an error
 *	Might depend on whether in relation to a previous-hop or a next-hop
 *	How does the user get informed so that something meaningful can be done
 *	How to do generic translation between ST errors and native (OS) errors
 *
 */

#define ErrorFlagList \
 anEFlg(FatalGroup,Error requires group to be removed) \
 anEFlg(FatalStream,Error requires stream to be removed) \
 anEFlg(FatalTarget,Error requires target to be removed) \
 anEFlg(FatalHop,Error requires all targets to nexthop to be removed) \
 anEFlg(FatalNet,Error requires all targets reached via net to be removed) \
 anEFlg(Reroute,Error requires rerouting) \
 anEFlg(Warning,No action required) \


enum errorflags {
    errorflag0 = 0,			/* zero is unspecified */
#define anEFlg(name,desc) name,
    ErrorFlagList
#undef anEFlg
};

/* ------------------------------------------------------------------------ */

/*
 *  ReasonCode
 *	Define ST-II Error Objects.  For each error: (internal) value, external
 *	name, internal name, control flags, and description.
 *
 *	ApplicationSpecificErrors is proviced to allow applications to
 *	insert additional (non-ST-II) elements in the enumeration, if desired.
 *
 */

#define ST2ErrorList \
 anError(0,NoError,NoError,Q(Warning) \
		,No error has been detected) \
 anError(1,ErrorUnknown,ErrorUnknown,Q(FatalStream) \
		,An error not contained in this list has been detected) \
 anError(2,AcceptTimeout,AcceptTimeout,Q(FatalTarget) \
		,An Accept has not been acknowledged) \
 anError(3,AccessDenied,AccessDenied,Q(FatalStream) \
		,Access denied) \
 anError(4,AckUnexpected,AckUnexpected,Q(Warning) \
		,An unexpected ACK was received) \
 anError(5,ApplAbort,ApplAbort,Q(FatalStream) \
		,The application aborted the stream abnormally) \
 anError(6,ApplDisconnect,ApplDisconnect,Q(FatalTarget) \
		,The application closed the stream normally) \
 anError(7,AuthentFailed,AuthentFailed,Q(FatalStream) \
		,The authentication function failed) \
 anError(8,CantGetResrc,CantGetResrc,Q(FatalTarget) \
		,Unable to acquire (additional) resources) \
 anError(9,CantRelResrc,CantRelResrc,Q(Warning) \
		,Unable to release excess resources) \
 anError(10,CksumBadCtl,CksumBadCtl,Q(Warning) \
		,A received control PDU has a bad message checksum) \
 anError(11,CksumBadST,CksumBadST,Q(Warning) \
		,A received PDU has a bad ST Header checksum) \
 anError(12,DropExcdDly,DropExcdDly,Q(Warning) \
		,A received PDU was dropped because it could not be processed within the delay specification) \
 anError(13,DropExcdMTU,DropExcdMTU,Q(Warning) \
		,A received PDU was dropped because its size exceeds the MTU) \
 anError(14,DropFailAgt,DropFailAgt,Q(Warning) \
		,A received PDU was dropped because of a failed ST agent) \
 anError(15,DropFailHst,DropFailHst,Q(Warning) \
		,A received PDU was dropped because of a host failure) \
 anError(16,DropFailIfc,DropFailIfc,Q(Warning) \
		,A received PDU was dropped because of a broken interface) \
 anError(17,DropFailNet,DropFailNet,Q(Warning) \
		,A received PDU was dropped because of a network failure) \
 anError(18,DropLimits,DropLimits,Q(Warning) \
		,A received PDU was dropped because it exceeds the resource limits for its stream) \
 anError(19,DropNoResrc,DropNoResrc,Q(Warning) \
		,A received PDU was dropped due to no available resources (including precedence)) \
 anError(20,DropNoRoute,DropNoRoute,Q(Warning) \
		,A received PDU was dropped because of no available route) \
 anError(21,DropPriLow,DropPriLow,Q(Warning) \
		,A received PDU was dropped because it has a priority too low to be processed) \
 anError(22,DuplicateIgn,DuplicateIgn,Q(Warning) \
		,A received control PDU is a duplicate and is being acknowledged) \
 anError(23,DuplicateTarget,DuplicateTarget,Q(Warning) \
		,A received control PDU contains a duplicate target or an attempt to add an existing target) \
 anError(24,FailureRecovery,FailureRecovery,Q(Warning) \
		,A notification that recovery is being attempted) \
 anError(25,FlowVerBad,FlowVerBad,Q(FatalStream) \
		,A received control PDU has a FlowSpec Version Number that is not supported) \
 anError(26,GroupUnknown,GroupUnknown,Q(Warning) \
		,A received control PDU contains an unknown Group Name) \
 anError(27,HIDDeferredMcst,HIDDeferredMcst,Q(Warning) \
		,Deferred HID selection specified with multicast group) \
 anError(28,HIDNegFails,HIDNegFails,Q(FatalHop) \
		,HID negotiation failed) \
 anError(29,HIDUnknown,HIDUnknown,Q(FatalHop) \
		,A received control PDU contains an unknown HID) \
 anError(30,InconsistHID,InconsistHID,Q(FatalHop) \
		,An inconsistency has been detected with a stream Name and corresponding HID) \
 anError(31,InconsistGroup,InconsistGroup,Q(Warning) \
		,An inconsistency has been detected with the streams forming a group) \
 anError(32,IntfcFailure,IntfcFailure,Q(FatalNet) \
		,A network interface failure has been detected) \
 anError(33,InvalidHID,InvalidHID,Q(FatalHop) \
		,A received ST PDU contains an invalid HID) \
 anError(34,InvalidSender,InvalidSender,Q(FatalStream) \
		,A received control PDU has an invalid SenderIPAddress field) \
 anError(35,InvalidTotByt,InvalidTotByt,Q(FatalStream) \
		,A received control PDU has an invalid TotalBytes field) \
 anError(36,LnkRefUnknown,LnkRefUnknown,Q(FatalHop) \
		,A received control PDU contains an unknown LnkReference) \
 anError(37,NameUnknown,NameUnknown,Q(FatalHop) \
		,A received control PDU contains an unknown stream Name (bad format?)) \
 anError(38,NetworkFailure,NetworkFailure,Q(FatalNet) \
		,A network failure has been detected) \
 anError(39,NoRouteToAgent,NoRouteToAgent,Q(Reroute) \
		,Cannot find a route to an ST agent) \
 anError(40,NoRouteToDest,NoRouteToDest,Q(Reroute) \
		,Cannot find a route to the destination) \
 anError(41,NoRouteToHost,NoRouteToHost,Q(Reroute) \
		,Cannot find a route to a host) \
 anError(42,NoRouteToNet,NoRouteToNet,Q(Reroute) \
		,Cannot find a route to a network) \
 anError(43,OpCodeUnknown,OpCodeUnknown,Q(FatalHop) \
		,A received control PDU has an invalid OpCode field) \
 anError(44,PCodeUnknown,PCodeUnknown,Q(FatalHop) \
		,A received control PDU has a parameter with an invalid PCode) \
 anError(45,ParmValueBad,ParmValueBad,Q(FatalHop) \
		,A received control PDU contains an invalid parameter value) \
 anError(46,PcolIdUnknown,PcolIdUnknown,Q(FatalTarget) \
		,A received control PDU contains an unknown next-higher layer protocol identifier) \
 anError(47,ProtocolError,ProtocolError,Q(FatalHop) \
		,A protocol error was detected) \
 anError(48,PTPError,PTPError,Q(FatalHop) \
		,Multiple targets were specified for a stream created with the PTP option) \
 anError(49,RefUnknown,RefUnknown,Q(FatalHop) \
		,A received control PDU contains an unknown Reference) \
 anError(50,RestartLocal,RestartLocal,Q(Reroute) \
		,The local ST agent has recently restarted) \
 anError(51,RemoteRestart,RemoteRestart,Q(Reroute) \
		,The remote ST agent has recently restarted) \
 anError(52,RetransTimeout,RetransTimeout,Q(FatalHop) \
		,An acknowledgment to a control message has not been received after several retransmissions) \
 anError(53,RouteBack,RouteBack,Q(Warning) \
		,The routing function indicates that the route to the next-hop is through the same interface as the previous-hop and is not the previous-hop) \
 anError(54,RouteInconsist,RouteInconsist,Q(Reroute) \
		,A routing inconsistency has been detected e.g. a route loop) \
 anError(55,RouteLoop,RouteLoop,Q(Reroute) \
		,A CONNECT was received that specified an existing target) \
 anError(56,SAPUnknown,SAPUnknown,Q(FatalTarget) \
		,A received control PDU contains an unknown next-higher layer SAP (port)) \
 anError(57,STAgentFailure,STAgentFailure,Q(FatalHop) \
		,An ST agent failure has been detected) \
 anError(58,StreamExists,StreamExists,Q(FatalHop) \
		,A stream with the given Name or HID already exists) \
 anError(59,StreamPreempted,StreamPreempted,Q(Reroute) \
		,The stream has been preempted by one with a higher precedence) \
 anError(60,STVerBad,STVerBad,Q(FatalHop) \
		,A received PDU is not ST Version 2) \
 anError(61,TooManyHIDs,TooManyHIDs,Q(FatalHop) \
		,Attempt to add more HIDs to a stream than the implementation supports) \
 anError(62,TruncatedCtl,TruncatedCtl,Q(FatalHop) \
		,A received control PDU is shorter than expected) \
 anError(63,TruncatedPDU,TruncatedPDU,Q(FatalHop) \
		,A received ST PDU is shorter than the ST Header indicates) \
 anError(64,UserDataSize,UserDataSize,Q(FatalHop) \
		,The UserData parameter is too large to permit a control message to fit into a networks MTU) \

#define ImplErrorList \
 anError(80,ParmValueBad,BadParmLen,Q(FatalHop) \
		,A parameter length not multiple of 4) \
 anError(81,ErrorUnknown,CantDeactivateActive,Q(Warning) \
		,An attempt to deactivate a connection which is still active) \
 anError(82,ParmValueBad,DupParm,Q(FatalHop) \
		,A control message contained multiple instances of a parameter) \
 anError(83,ProtocolError,HelloTimerandDetectorIPAdr,Q(FatalHop) \
		,A pdu cannot have both a HelloTimer and a DetectorIPAdr) \
 anError(84,ProtocolError,HIDandReasonCode,Q(FatalHop) \
		,A pdu cannot have both a HID and a ReasonCode) \
 anError(85,ProtocolError,ImplRest,Q(FatalHop) \
		,Attempt to exceed capabilities of the implementation) \
 anError(86,ErrorUnknown,Inconsistency,Q(FatalHop) \
		,An unexpected condition has occurred) \
 anError(87,InconsistHID,InconsistVLID,Q(FatalHop) \
		,An inconsistency has been detected with a stream Name and corresponding VLID) \
 anError(88,ProtocolError,InvalidArgument,Q(Warning) \
		,A subroutine was called with an invalid argument) \
 anError(89,ErrorUnknown,InvalidPointer,Q(Warning) \
		,An attempt to use an undefined pointer has occurred) \
 anError(90,ProtocolError,InvalidRVLId,Q(FatalHop) \
		,A received SCMP PDU contains an invalid RVLId) \
 anError(91,ProtocolError,InvalidSVLId,Q(FatalHop) \
		,A received SCMP PDU contains an inconsistent SVLId) \
 anError(92,ProtocolError,InvalidTarget,Q(FatalTarget) \
		,A Target is improperly specified) \
 anError(93,ProtocolError,MTUtosmall,Q(FatalHop) \
		,A network MTU is too small for SCMP messages) \
 anError(94,CantGetResrc,NoConHidVlid,Q(FatalHop) \
		,Unable to acquire a connection block or HID or VLId) \
 anError(95,ErrorUnknown,NoSupport,Q(FatalHop) \
		,Attempt to use non-supported feature) \
 anError(96,ProtocolError,ParmMissing,Q(FatalHop) \
		,A mandatory parameter is not present) \
 anError(97,ProtocolError,SourceRouteError,Q(FatalTarget) \
		,A Target has an invalid source route) \
 anError(98,CantGetResrc,InvalidRsrcId,Q(FatalHop) \
		,Use of an invalid resource identifier) \
 anError(99,ErrorUnknown,Misconfiguration,Q(FatalStream) \
		,A local configuration error has been detected) \
 anError(100,NoError,InProgress,Q(Warning) \
		,operation in progress) \
 anError(101,ErrorUnknown,Alignment,Q(Warning) \
		,A memory object is not correctly aligned) \
 anError(102,ErrorUnknown,EncapsulationError,Q(FatalHop) \
		,Error processing encapsulated ST) \


#ifndef ApplicationSpecificErrors
#define ApplicationSpecificErrors
#endif

enum ST2Errors {
#define anError(vi,vx,n,f,def) n = vi,
    ST2ErrorList

    NProtocolErrors,		/* gets value last+1, # formal errors */

/* Following are not in protocol spec */

    ImplErrorList

    ApplicationSpecificErrorBase = 256,

    ApplicationSpecificErrors
#undef anError

 /*  MAX_ST_ERRORS		/* has value last+1, # implementation errors */
#define MAX_ST_ERRORS (127)	/* C won't use an enum to declare array */
				/* sizes! */
};


/*
C* ST-II Configuration Definitions
c*
c*	The ConfigParmList defining site's configuration parameters is
c*	instantiated in st2_proto so sites can make local additions.
c*	Only add to the end of the list unless you recompile all of
c*	ST-II source!
c*
c*	Placed here so Applications can get at the information, too.
c*
c*	WARNING: Exercise caution when changing the definitions.
c*
 *	name,current value,min value,max value,compile symbol,format,desc
 *	aCfgParm(name,cv,nv,xv,cs,format,desc)
 *
 *	format specifies how to print the value as "i" quantities of "ni"
 *	bytes each (i < 5) as <i><n1><n2>...<ni><printf format specification>
 *	Currently, the <printf format specification> prints 13 characters.
 */


/*	ConfigVersion	(3)		/* Source         1.9 */
#define ConfigVersion	(4)		/* Source Release 1.10 */


#define ConfigParm(x)	st2_config.x

#define ConfigParmList \
 aCfgParm(size,sizeof (struct aConfigBlock),8,, \
		,"14%17u.",size of this block) \
 aCfgParm(version,Experimenter|ConfigVersion,,, \
		,"14%17x ",Experimenter|ConfigVersion version of this block) \
	\
 aCfgParm(api_maxctl,MCLBYTES,128,xx,MCLBYTES \
		,"14%17u.",MTU through API to applications) \
 aCfgParm(api_maxnam,MLEN,16,xx,MLEN \
		,"14%17u.",MTU through API to applications) \
 aCfgParm(api_MTU,MCLBYTES,128,xx,MCLBYTES \
		,"14%17u.",MTU through API to applications) \
 aCfgParm(coip_family,AF_COIP,AF_COIP,AF_COIP,AF_COIP \
		,"14%17u.",Address family for COIP) \
 aCfgParm(ctlflg,CTLFLG,0,0xffffffff, \
		,"14%17x ",ST2Flags) \
 aCfgParm(dbgflg,DBGFLG,0,0xffffffff, \
		,"14%17x ",Control diagnostic output) \
 aCfgParm(def_genifs,DEF_GENIFS,2,(#nets+MAX_STREAMS),DEF_GENIFS \
		,"14%17u.",# nets + Origins + Targets) \
 aCfgParm(def_nxthops,DEF_NXTHOPS,1,xx,DEF_NXTHOPS \
		,"14%17u.",# nxthops per stream w/o ExpAry) \
 aCfgParm(def_routechoices,2,1,xx,DEF_ROUTECHOICES \
		,"14%17u.",estimated # routes to a destination) \
 aCfgParm(ethertype,0x5354,1501,0xFFFF, \
		,"14%17x ",private ethertype when x800 cannot be used) \
 aCfgParm(flowspec_versions,0,0,0xFFFFFFFF, \
		,"14%17x ",bitmap of supported flowspec versions) \
 aCfgParm(fhid_masks,FHID_MASKS,1,62,FHID_MASKS \
		,"14%17u.",RFC1190 recommends at least 64 bits) \
 aCfgParm(hellobackoff,20000,1,0xFFFFFFFF,nodefine \
		,"14%17u.",HELLO interval backoff msec) \
 aCfgParm(hellofiltfactor,2,1,xx,nodefine \
		,"14%17u.",HELLO filtering factor) \
 aCfgParm(hellolossfactor,5,1,xx,nodefine \
		,"14%17u.",HELLO loss  factor) \
 aCfgParm(hellotimerholddown,10000,1000,60000,nodefine \
		,"14%17u.",restart quiet time in msec) \
 aCfgParm(hid_bits,HID_BITS,3,16,ID_BITS \
		,"14%17u.",# of LSB HID bits we use) \
 aCfgParm(hid_mask,NBitMask (HID_BITS),7,0xFFFF, \
		,"14%17x ",mask for LSB HID bits we use) \
 aCfgParm(ip_tos,0x00380038,0,0x00FE00FE,nodefine \
		,"222 %7x  %7x ",IP TOS neighbor/stream) \
 aCfgParm(ip_ttl,0x00400040,0,0x00FF00FF,nodefine \
		,"222 %7u. %7u.",IP TTL neighbor/stream) \
 aCfgParm(max_hids_per_stream,MAX_HIDS_PER_STREAM,2,xx,MAX_HIDS_PER_STREAM \
		,"14%17u.",) \
 aCfgParm(max_lclhdrlen,MAX_LCLHDRLEN,0,999,MAX_LCLHDRLEN \
		,"14%17u.",maximum local network header length) \
 aCfgParm(max_st_errors,127,xx,xx,MAX_ST_ERRORS \
		,"14%17u.",) \
 aCfgParm(max_streams,MAX_STREAMS,3,65536,MAX_STREAMS \
		,"14%17u.",max simultaneous streams + 2) \
 aCfgParm(n32freehids,(((1 << HID_BITS) + 31) / 32),1,2084, \
		,"14%17u.",# 32-bit words for (1 << HID_BITS) bits) \
 aCfgParm(nbrRTT0,4000,1,0xFFFFFFFF,nodefine \
		,"14%17u.",initial value of RTT to a neighbor) \
 aCfgParm(noc0,0,0,0xFFFFFFFF,nodefine \
		,"41111     %u.%u.%u.%u",primary NOC) \
 aCfgParm(noc1,0,0,0xFFFFFFFF,nodefine \
		,"41111     %u.%u.%u.%u",backup NOC) \
 aCfgParm(noc2,0,0,0xFFFFFFFF,nodefine \
		,"41111     %u.%u.%u.%u",second backup NOC) \
 aCfgParm(ntraps,MAX_TRAPS,1,10000,MAX_TRAPS \
		,"14%17u.",number of entries in traptable) \
 aCfgParm(num_nexthop,100,xx,xx,NUM_NEXTHOP \
		,"14%17u.",) \
 aCfgParm(num_st_opcodes,NUM_ST_OPCODES,NUM_ST_OPCODES,NUM_ST_OPCODES,NUM_ST_OPCODES \
		,"14%17u.",# of OpCodes) \
 aCfgParm(num_st_parms,NUM_ST_PARMS,NUM_ST_PARMS,NUM_ST_PARMS,NUM_ST_PARMS \
		,"14%17u.",# of Parameters) \
 aCfgParm(rcvryhold,10000,2500,60000,nodefine \
		,"14%17u.",(notyet)msec to hold streams pending reconnection after failure) \
 aCfgParm(rcvrytodef,10000,2500,60000,DEF_RCVRYTO \
		,"14%17u.",Default msec to detect failed path) \
 aCfgParm(rcvrytomin,5000,2500,60000,MIN_RCVRYTO \
		,"14%17u.",minimum msec to detect failed path) \
 aCfgParm(rcvrytonbr,100000,15000,60000,NBR_RCVRYTO \
	        ,"14%17u.",msec to detect failed neighbor when no active streams) \
 aCfgParm(recvspace,1024*4,1024,64*1024, \
		,"14%17u.",default socket recv space) \
	/* BEGIN following are each struct stRetryParms */ \
 aCfgParm(rx_Accept,(3<<16)+1000,xx,xx,xx \
		,"222 %7u. %7u.",ACCEPT retransmission count & interval) \
 aCfgParm(rx_Ack,(0<<16)+0000,xx,xx,xx \
		,"222 %7u. %7u.",ACK retransmission count & interval) \
 aCfgParm(rx_Change,(3<<16)+1000,xx,xx,xx \
		,"222 %7u. %7u.",CHANGE retransmission count & interval) \
 aCfgParm(rx_ChangeRequest,(3<<16)+1000,xx,xx,xx \
		,"222 %7u. %7u.",CHANGEREQUEST retransmission count & interval) \
 aCfgParm(rx_Connect,(5<<16)+1000,xx,xx,xx \
		,"222 %7u. %7u.",CONNECT retransmission count & interval) \
 aCfgParm(rx_Disconnect,(3<<16)+1000,xx,xx,xx \
		,"222 %7u. %7u.",DISCONNECT retransmission count & interval) \
 aCfgParm(rx_End2End,(0<<16)+5000,xx,xx,xx \
		,"222 %7u. %7u.",END2END retransmission count & interval) \
 aCfgParm(rx_ErrorInRequest,(0<<16)+0000,xx,xx,xx \
		,"222 %7u. %7u.",ERRORINREQUEST retransmission count & interval) \
 aCfgParm(rx_ErrorInResponse,(0<<16)+0000,xx,xx,xx \
		,"222 %7u. %7u.",ERRORINRESPONSE retransmission count & interval) \
 aCfgParm(rx_Hello,(0<<16)+0,xx,xx,xx \
		,"222 %7u. %7u.",HELLO retransmission count & interval) \
 aCfgParm(rx_HIDApprove,(0<<16)+0000,xx,xx,xx \
		,"222 %7u. %7u.",HIDAPPROVE retransmission count & interval) \
 aCfgParm(rx_HIDChange,(3<<16)+1000,xx,xx,xx \
		,"222 %7u. %7u.",HIDCHANGE retransmission count & interval) \
 aCfgParm(rx_HIDChangeRequest,(3<<16)+1000,xx,xx,xx \
		,"222 %7u. %7u.",HIDCHANGEREQUEST retransmission count & interval) \
 aCfgParm(rx_HIDReject,(0<<16)+0000,xx,xx,xx \
		,"222 %7u. %7u.",HIDREJECT retransmission count & interval) \
 aCfgParm(rx_Notify,(3<<16)+1000,xx,xx,xx \
		,"222 %7u. %7u.",NOTIFY retransmission count & interval) \
 aCfgParm(rx_Refuse,(3<<16)+1000,xx,xx,xx \
		,"222 %7u. %7u.",REFUSE retransmission count & interval) \
 aCfgParm(rx_Reroute,(5<<16)+1000,xx,xx,xx \
		,"222 %7u. %7u.",Reroute retransmission count & interval) \
 aCfgParm(rx_Status,(3<<16)+1000,xx,xx,xx \
		,"222 %7u. %7u.",STATUS retransmission count & interval) \
 aCfgParm(rx_StatusResponse,(0<<16)+0,xx,xx,xx \
		,"222 %7u. %7u.",STATUSRESPONSE retransmission count & interval) \
	/* END struct stRetryParms */ \
 aCfgParm(sendspace,1024*4,1024,64*1024, \
		,"14%17u.",default socket send space) \
 aCfgParm(sweep_interval,20,1,86400, \
		,"14%17u.",ticks between checking for hung streams) \
 aCfgParm(targ_lists,TARG_LISTS,1,xx,TARG_LISTS \
		,"14%17u.",# target lists to encode per SCMP) \
 aCfgParm(targ_per_list,TARG_PER_LIST,1,xx,TARG_PER_LIST \
		,"14%17u.",# targets to encode per target list) \
 aCfgParm(targ_per_scmp_def,TARG_PER_SCMP_DEF,1,xx,TARG_PER_SCMP_DEF \
		,"14%17u.",# targets parsed w/o ExpAry) \
 aCfgParm(timeoutfactor,1,1,100, \
		,"14%17u.",retransmit timeout factor - slowdown for debug) \
 aCfgParm(vlnk_bits,VLNK_BITS,1,VLID_BITS,VLNK_BITS \
		,"14%17u.",number of VLID_BITS used for table index) \
 aCfgParm(vlnk_hash,VLNK_HASH,1,VLNK_HASH,VLNK_HASH \
		,"14%17u.",number of entries in (remote) VLink hash table) \
 aCfgParm(vlnk_mask,((1<<VLNK_BITS)-1),1,NBitMask (VLNK_BITS),VLNK_MASK \
		,"14%17x ",mask for index bits in a VLID) \


struct aConfigBlock {
    unsigned long
#define aCfgParm(name,cv,nv,xv,cs,fo,desc) name,
	ConfigParmList
#undef aCfgParm
			 end;
};

extern struct aConfigBlock st2_config;

/* ------------------------------------------------------------------------ */

/* ST-II Implementation Control Flags.
 *
 */

#define ST2FlagList \
 aST2Flg(0x00000001,ST2FlgNoSrcChking \
			,Do not record and check source of ST-II packets) \
 aST2Flg(0x00000080,ST2FlgNoHello \
			,Do not send HELLOs) \
 aST2Flg(0x00000100,ST2FlgNoFail \
			,Do not abort streams on neighbor failure) \
 aST2Flg(0x20000000,ST2FlgSTCksm \
			,Accept ST packets with a zero ckecksum) \
 aST2Flg(0x40000000,ST2FlgSCMPCksm,Accept SCMP packets with a zero ckecksum) \
 aST2Flg(0x80000000,ST2FlgBugPrint,Print/Log BUGCHECK info) \


#define ConfigFlag(x) ((st2_config.ctlflg & (unsigned long) (x)) NE 0)


enum ST2Flag {
#define aST2Flg(bit,name,desc) name = bit,
    ST2FlagList
#undef aST2Flg
    ST2FlgDummy
};


/*
 * Service costs.
 */

#define CostList \
 aCostVar(ctrl,"14%17u.",) \
 aCostVar(ctrl_drop,"14%17u.",) \
 aCostVar(data,"14%17u.",) \
 aCostVar(data_drop,"14%17u.",) \
 aCostVar(fixed,"14%17u.",) \
 aCostVar(os,"14%17u.",) \


/* ------------------------------------------------------------------------ */

/*
 * Packet throughput counters.
 */

#define ThruList \
 aThruVar(ctrl_in_bytes,"14%17u.",) /* NB: must be first */ \
 aThruVar(ctrl_in_pkts,"14%17u.",) \
 aThruVar(ctrl_in_drop_bytes,"14%17u.",) \
 aThruVar(ctrl_in_drop_pkts,"14%17u.",) \
 aThruVar(ctrl_out_bytes,"14%17u.",) \
 aThruVar(ctrl_out_pkts,"14%17u.",) \
 aThruVar(ctrl_out_drop_bytes,"14%17u.",) \
 aThruVar(ctrl_out_drop_pkts,"14%17u.",) \
 aThruVar(data_in_bytes,"14%17u.",) \
 aThruVar(data_in_pkts,"14%17u.",) \
 aThruVar(data_in_drop_bytes,"14%17u.",) \
 aThruVar(data_in_drop_pkts,"14%17u.",) \
 aThruVar(data_out_bytes,"14%17u.",) \
 aThruVar(data_out_pkts,"14%17u.",) \
 aThruVar(data_out_drop_bytes,"14%17u.",) \
 aThruVar(data_out_drop_pkts,"14%17u.",) \



/*
A* ST-II Statistics Definitions
a*
a*	The CounterList is instantiated in st2_proto so sites can make
a*	local additions.  Only add to the end of the list unless you
a*	recompile all of ST-II source!
a*
a*	Placed here so Applications can get at the information, too.
a*
 *	name,format,description
 *	aCounter(name,format,description)
 *
 *	format specifies how to print the value as "i" quantities of "ni"
 *	bytes each (i < 5) as <i><n1><n2>...<ni><printf format specification>
 *	Currently, the <printf format specification> prints 18 characters.
 *
 */

/*      ST2StatsVersion (1)		/* Binary Release 1 */
/*	ST2StatsVersion (2)		/* Source Release 1.8 */
/*	ST2StatsVersion (3)		/* Source         1.9  */
#define ST2StatsVersion (4)		/* Source Release 1.10  */


#define CounterList \
	/* WARNING, Start: Do NOT change this block */ \
 aCounter(length,"14%17u.",)		/* sizeof (st2_stats), not counter */ \
 aCounter(version,"14%17u.",)		/* version of statistics block */ \
 aCounter(agent,"41111   %u.%u.%u.%u",)	/* IPAdr of the agent */ \
 aCounter(rptSeconds,"244%10u.%06u ",)	/* Timestamp of data, seconds */ \
 aCounter(rptFract,"0",)		/* ... fractional seconds */ \
 aCounter(butSeconds,"244%10u.%06u ",)	/* Timestamp when booted, seconds */ \
 aCounter(butFract,"0",)		/* ... fractional seconds */ \
 aCounter(unique_id,"14%17u.",)		/* Unique Id */ \
 aCounter(other,"14%17u.",)		/* Catch-all */ \
	/* WARNING, End: Do NOT change this block */ \
 aCounter(acksdelayed,"14%17u.",) \
 aCounter(bad_mt_type,"14%17u.",) \
 aCounter(bad_stcksum,"14%17u.",) \
 aCounter(bug_checks,"14%17u.",) \
 aCounter(bug_id,"14%17x ",) \
 aCounter(bug_info,"14%17x ",) \
    CostList \
    ThruList \
 aCounter(encodebadparm,"14%17u.",) \
 aCounter(encodewrongparms,"14%17u.",) \
 aCounter(errfreeingcon,"14%17u.",) \
 aCounter(errfreeinghid,"14%17u.",) \
 aCounter(errfreeingvlid,"14%17u.",) \
 aCounter(fasttics,"14%17u.",) \
 aCounter(frwd_nonxthops,"14%17u.",) \
 aCounter(frwd_notclmbuf,"14%17u.",) \
 aCounter(hellomsec,"14%17u.",our HelloTimer) \
 aCounter(hellorecv,"14%17u.",) \
 aCounter(hellosent,"14%17u.",) \
 /* Begin keep together - struct timeval */ \
 aCounter(hellosec,"14%17u.",local tv_sec of HelloTimer) \
 aCounter(hellousec,"14%17u.",local tv_usec of HelloTimer) \
 /* End keep together */ \
 aCounter(lennot4x_parm,"14%17u.",) \
 aCounter(lennot4x_target,"14%17u.",) \
 aCounter(misconfiguration,"14%17u.",) \
 aCounter(msg_inhidcoll,"14%17u.",) \
 aCounter(msg_innocon,"14%17u.",) \
 aCounter(msg_intrunc,"14%17u.",) \
 aCounter(nbrsactivated,"14%17u.",# neighbors started sending HELLOs to) \
 aCounter(nbrscreated,"14%17u.",# neighbors created) \
 aCounter(nbrsdeactivated,"14%17u.",# neighbors stopped sending HELLOs to) \
 aCounter(nbrsdeleted,"14%17u.",# neighbors deleted) \
 aCounter(nbrsfailed,"14%17u.",# neighbor instanced failed via HELLOs) \
 aCounter(nbrsfailedRemoteRestart,"14%17u.",# neighbor instanced failed: RemoteRestart) \
 aCounter(nbrsfailedstream,"14%17u.",# neighbor instanced failed via HELLOs) \
 aCounter(nbrsfailedtarget,"14%17u.",# neighbor instanced failed via HELLOs) \
 aCounter(nbrsinserted,"14%17u.",# vlinks monitored to all neighbors) \
 aCounter(nbrsremoved,"14%17u.",# vlinks no longer monitored to all neighbors) \
 aCounter(no_dspin,"14%17u.",) \
 aCounter(no_ifnet,"14%17u.",) \
 aCounter(no_mt_data,"14%17u.",) \
 aCounter(nofreecon,"14%17u.",) \
 aCounter(nofreehids,"14%17u.",) \
 aCounter(nofreevlids,"14%17u.",) \
 aCounter(nofwdifc,"14%17u.",) \
 aCounter(not_aligned,"14%17u.",) \
 aCounter(not_st2,"14%17u.",) \
 aCounter(parm_missing,"14%17u.",) \
	\
 aCounter(net_in_bytes,"14%17u.",) \
 aCounter(net_in_pkts,"14%17u.",) \
 aCounter(net_in_drop_bytes,"14%17u.",) \
 aCounter(net_in_drop_pkts,"14%17u.",) \
 aCounter(net_out_bytes,"14%17u.",) \
 aCounter(net_out_pkts,"14%17u.",) \
 aCounter(net_out_drop_bytes,"14%17u.",) \
 aCounter(net_out_drop_pkts,"14%17u.",) \
	\
 aCounter(pcb_fail,"14%17u.",pcbs not allocated -- aka nofreecon) \
 aCounter(pcb_get,"14%17u.",pcbs allocated) \
 aCounter(pcb_rel,"14%17u.",pcbs released) \
 aCounter(pkts_in,"14%17u.",) \
 aCounter(pkts_in_drop,"14%17u.",) \
 aCounter(pkts_in_ipv4p,"14%17u.",) \
 aCounter(pkts_in_ipv4p5,"14%17u.",) \
 aCounter(pkts_in_ipv4p5_rawip,"14%17u.",) \
 aCounter(pkts_in_ipv4px,"14%17u.",) \
 aCounter(pkts_in_ipv5,"14%17u.",) \
 aCounter(pkts_in_max,"14%17u.",input queue high water mark) \
 aCounter(pkts_in_qfull,"14%17u.",input dropped by net because queue full) \
 aCounter(sap_next,"14%17u.",) \
 aCounter(scmp_0rvlid,"14%17u.",) \
 aCounter(scmp_0svlid,"14%17u.",) \
 aCounter(scmp_badvlid,"14%17u.",) \
 aCounter(scmp_cksum,"14%17u.",) \
 aCounter(scmp_failed_ifc,"14%17u.",) \
 aCounter(scmp_failed_neighbor,"14%17u.",) \
 aCounter(scmp_failed_net,"14%17u.",) \
 aCounter(scmp_failed_resources,"14%17u.",) \
 aCounter(scmp_failed_route,"14%17u.",) \
 aCounter(scmp_failed_route_fixed,"14%17u.",) \
 aCounter(scmp_failed_srcrut,"14%17u.",) \
 aCounter(scmp_failed_to_route,"14%17u.",) \
 aCounter(scmp_fragmented,"14%17u.",) \
 aCounter(scmp_leninconsist,"14%17u.",) \
 aCounter(scmp_nobfxhd,"14%17u.",insufficient packet header space) \
 aCounter(scmp_nobfxo1,"14%17u.",insufficient pre-overhead space) \
 aCounter(scmp_nobfxo2,"14%17u.",insufficient post-overhead space) \
 aCounter(scmp_nobfxpk,"14%17u.",insufficient packet space) \
 aCounter(scmp_nobuf0,"14%17u.",no buf to use) \
 aCounter(scmp_nobuf1,"14%17u.",no small bufs) \
 aCounter(scmp_nobuf2,"14%17u.",no large bufs) \
 aCounter(scmp_nobuf3,"14%17u.",no buf big enough) \
 aCounter(scmp_nobufmod,"14%17u.",invalid GetSpace mode) \
 aCounter(scmp_refnum,"14%17u.",) \
 aCounter(scmp_rx,"14%17u.",number of retransmitted control messages) \
 aCounter(scmp_rx_timeout,"14%17u.",number of unacknowledged control messages) \
 aCounter(scmp_tooshort,"14%17u.",) \
 aCounter(slowtics,"14%17u.",) \
 aCounter(trapcount,"14%17u.",number of traps) \
 aCounter(trapnext,"14%17x ",next free slot in traptable) \
 aCounter(trappage,"14%17u.",) \
 aCounter(trg_fail,"14%17u.",targets not allocated) \
 aCounter(trg_get,"14%17u.",targets allocated) \
 aCounter(trg_rel,"14%17u.",targets released) \
 aCounter(vlinkfail,"14%17u.",) \
 aCounter(vlinkget,"14%17u.",) \
 aCounter(vlinkrel,"14%17u.",) \

struct aStatsBlock {
#define aCostVar(name,format,desc) unsigned long Ident(cost_)name;
#define aCounter(name,format,desc) unsigned long name;
#define aThruVar(name,format,desc) unsigned long name;
    CounterList
#undef  aCostVar
#undef  aCounter
#undef  aThruVar
};

extern struct aStatsBlock	st2_stats;


/* Macro to increment a counter */
#define Count(n,v)	st2_stats.n += (v)



/* ??? Define network management access functions, including accounting. */

/*
T* Testing and Debugging Print Routines
t*
t*	The ST2Print Module provides routines to print ST-II Protocol
t*	Data Units (PrintXxx) and implementation data structures
t*	(print_xxx).  The subroutine declarations are availble to
t*	Applications when USING_ST2Print is defined before including
t*	st2_api h.
t*
 */

#ifdef USING_ST2Print

extern void	 flush_history (/* */);
extern FILE	*g_fp;			/* initialized to stdout */
extern char	*LookupSTError (/* enum ST2Errors error */);
extern char	*LookupSTOpcode (/* enum anSTOpCode, char *default */);
extern void	 print_aNeighborList (/* prefix */);
extern int	 print_aNetIf (/* kagifca, agifcp, prefix */);
extern void	 print_aNxtHop (/* long anha, prefix */);
extern void	 print_aNxtHopList (/* long anhla, prefix */);
extern void	 print_aPktDesc (/* long apda, prefix */);
extern void	 print_aRouteChoiceList (/* arcla, arclp, prefix */);
extern int	 print_arptable (/* prefix */);
extern void	 print_aRsrcBuf (/* long arba, prefix */);
extern void	 print_aRsrcCPU (/* long arca, prefix */);
extern void	 print_aRsrcNet (/* long arna, prefix */);
extern int	 print_aST2pcb (/* long kcp, prefix */);
extern void	 print_aTarget (/* long ata, atl, prefix */);
extern void	 print_aVLink (/* long vla, prefix */);
extern void	 print_config (/* prefix */);
extern int	 print_ConHshTable (/* prefix */);
extern void	 print_flow_spec (/* long fsa, prefix */);
extern int	 print_gifcs (/* prefix */);
extern int	 print_netrsrcs (/* prefix */);
extern void	 print_next_hop (/* long nha, prefix */);
extern void	 print_route (/* long rp, prefix */);
extern int	 print_rxqueue (/* prefix */);
extern void	 print_socket (/* long sop, prefix */);
extern int	 print_stats (/* prefix */);
extern void	 print_traps (/* prefix */);
extern void	 print_vlinks (/* prefix */);
extern int	 PrintIP4Header (/* ipp, len, prefix */);
extern int	 PrintMsg (/* struct msghdr *msgp, datalen, prefix */);
extern int	 PrintParms (/* struct aParameter *parmp, len, prefix */);
extern int	 PrintSockaddrST (/* struct sockaddr_st2 *sap, len, prefix */);
extern int	 PrintSCMPHeader (/* struct aSCMPHeader *scmhp, len, prefix */);
extern int	 PrintST2Header (/* struct aST2Header *st2hp, len, prefix */);
extern int	 PrintTarget (/* struct aTarget *tp, len, prefix */);

#endif USING_ST2Print


#endif  _ST2_API_H_
